#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009, 2012, 2013, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Esben S. Nielsen <esn@dmi.dk>

# This file is part of the mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the image class. It overlaps largely the PIL library,
but has the advandage of using masked arrays as pixel arrays, so that data
arrays containing invalid values may be properly handled.
"""
import warnings
warnings.warn(__name__ + " is deprecated, please use trollimage instead.",
              DeprecationWarning)

import os
import re

from PIL import Image as Pil
import numpy as np

try:
    import numexpr as ne
except ImportError:
    ne = None

from mpop.utils import ensure_dir

import logging

LOG = logging.getLogger(__name__)


class UnknownImageFormat(Exception):

    """Exception to be raised when image format is unknown to MPOP"""
    pass


def check_image_format(fformat):
    cases = {"jpg": "jpeg",
             "jpeg": "jpeg",
             "tif": "tiff",
             "tiff": "tif",
             "pgm": "ppm",
             "pbm": "ppm",
             "ppm": "ppm",
             "bmp": "bmp",
             "dib": "bmp",
             "gif": "gif",
             "im": "im",
             "pcx": "pcx",
             "png": "png",
             "xbm": "xbm",
             "xpm": "xpm",
             }
    fformat = fformat.lower()
    try:
        fformat = cases[fformat]
    except KeyError:
        raise UnknownImageFormat("Unknown image format '%s'." % fformat)
    return fformat

try:
    from numpy import percentile
except ImportError:

    # Stolen from numpy 1.7.0 for backward compatibility

    def percentile(a, q, axis=None, out=None, overwrite_input=False):
        """
        Compute the qth percentile of the data along the specified axis.

        Returns the qth percentile of the array elements.

        Parameters
        ----------
        a : array_like
        Input array or object that can be converted to an array.
        q : float in range of [0,100] (or sequence of floats)
        Percentile to compute which must be between 0 and 100 inclusive.
        axis : int, optional
        Axis along which the percentiles are computed. The default (None)
        is to compute the median along a flattened version of the array.
        out : ndarray, optional
        Alternative output array in which to place the result. It must
        have the same shape and buffer length as the expected output,
        but the type (of the output) will be cast if necessary.
        overwrite_input : bool, optional
        If True, then allow use of memory of input array `a` for
        calculations. The input array will be modified by the call to
        median. This will save memory when you do not need to preserve
        the contents of the input array. Treat the input as undefined,
        but it will probably be fully or partially sorted.
        Default is False. Note that, if `overwrite_input` is True and the
        input is not already an array, an error will be raised.

        Returns
        -------
        pcntile : ndarray
        A new array holding the result (unless `out` is specified, in
        which case that array is returned instead). If the input contains
        integers, or floats of smaller precision than 64, then the output
        data-type is float64. Otherwise, the output data-type is the same
        as that of the input.

        See Also
        --------
        mean, median

        Notes
        -----
        Given a vector V of length N, the qth percentile of V is the qth ranked
        value in a sorted copy of V. A weighted average of the two nearest
        neighbors is used if the normalized ranking does not match q exactly.
        The same as the median if ``q=50``, the same as the minimum if ``q=0``
        and the same as the maximum if ``q=100``.

        Examples
        --------
        >>> a = np.array([[10, 7, 4], [3, 2, 1]])
        >>> a
        array([[10, 7, 4],
        [ 3, 2, 1]])
        >>> np.percentile(a, 50)
        3.5
        >>> np.percentile(a, 0.5, axis=0)
        array([ 6.5, 4.5, 2.5])
        >>> np.percentile(a, 50, axis=1)
        array([ 7., 2.])

        >>> m = np.percentile(a, 50, axis=0)
        >>> out = np.zeros_like(m)
        >>> np.percentile(a, 50, axis=0, out=m)
        array([ 6.5, 4.5, 2.5])
        >>> m
        array([ 6.5, 4.5, 2.5])

        >>> b = a.copy()
        >>> np.percentile(b, 50, axis=1, overwrite_input=True)
        array([ 7., 2.])
        >>> assert not np.all(a==b)
        >>> b = a.copy()
        >>> np.percentile(b, 50, axis=None, overwrite_input=True)
        3.5

        """
        a = np.asarray(a)

        if q == 0:
            return a.min(axis=axis, out=out)
        elif q == 100:
            return a.max(axis=axis, out=out)

        if overwrite_input:
            if axis is None:
                sorted = a.ravel()
                sorted.sort()
            else:
                a.sort(axis=axis)
                sorted = a
        else:
            sorted = np.sort(a, axis=axis)
        if axis is None:
            axis = 0

        return _compute_qth_percentile(sorted, q, axis, out)

    # handle sequence of q's without calling sort multiple times
    def _compute_qth_percentile(sorted, q, axis, out):
        if not np.isscalar(q):
            p = [_compute_qth_percentile(sorted, qi, axis, None)
                 for qi in q]

            if out is not None:
                out.flat = p

            return p

        q = q / 100.0
        if (q < 0) or (q > 1):
            raise ValueError("percentile must be either in the range [0,100]")

        indexer = [slice(None)] * sorted.ndim
        Nx = sorted.shape[axis]
        index = q * (Nx - 1)
        i = int(index)
        if i == index:
            indexer[axis] = slice(i, i + 1)
            weights = np.array(1)
            sumval = 1.0
        else:
            indexer[axis] = slice(i, i + 2)
            j = i + 1
            weights = np.array([(j - index), (index - i)], float)
            wshape = [1] * sorted.ndim
            wshape[axis] = 2
            weights.shape = wshape
            sumval = weights.sum()

        # Use add.reduce in both cases to coerce data type as well as
        # check and use out array.
        return np.add.reduce(sorted[indexer] * weights, axis=axis, out=out) / sumval


class Image(object):

    """This class defines images. As such, it contains data of the different
    *channels* of the image (red, green, and blue for example). The *mode*
    tells if the channels define a black and white image ("L"), an rgb image
    ("RGB"), an YCbCr image ("YCbCr"), or an indexed image ("P"), in which case
    a *palette* is needed. Each mode has also a corresponding alpha mode, which
    is the mode with an "A" in the end: for example "RGBA" is rgb with an alpha
    channel. *fill_value* sets how the image is filled where data is missing,
    since channels are numpy masked arrays. Setting it to (0,0,0) in RGB mode
    for example will produce black where data is missing."None" (default) will
    produce transparency (thus adding an alpha channel) if the file format
    allows it, black otherwise.

    The channels are considered to contain floating point values in the range
    [0.0,1.0]. In order to normalize the input data, the *color_range*
    parameter defines the original range of the data. The conversion to the
    classical [0,255] range and byte type is done automagically when saving the
    image to file.
    """
    modes = ["L", "LA", "RGB", "RGBA", "YCbCr", "YCbCrA", "P", "PA"]

    def __init__(self, channels=None, mode="L", color_range=None,
                 fill_value=None, palette=None):

        self.channels = None
        self.mode = None
        self.width = 0
        self.height = 0
        self.fill_value = None
        self.palette = None
        self.shape = None
        self.info = {}

        self._secondary_mode = "RGB"

        if(channels is not None and
           not isinstance(channels, (tuple, set, list,
                                     np.ndarray, np.ma.core.MaskedArray))):
            raise TypeError("Channels should a tuple, set, list, numpy array, "
                            "or masked array.")

        if(isinstance(channels, (tuple, list)) and
           len(channels) != len(re.findall("[A-Z]", mode))):
            raise ValueError("Number of channels does not match mode.")

        if mode not in self.modes:
            raise ValueError("Unknown mode.")

        if(color_range is not None and
           not _is_pair(color_range) and
           not _is_list_of_pairs(color_range)):
            raise ValueError("Color_range should be a pair"
                             " or a list/tuple/set of pairs.")
        if(color_range is not None and
           _is_list_of_pairs(color_range) and
           (channels is None or
                len(color_range) != len(channels))):
            raise ValueError("Color_range length does not match number of "
                             "channels.")

        if(color_range is not None and
           (((mode == "L" or mode == "P") and not _is_pair(color_range)) and
                (len(color_range) != len(re.findall("[A-Z]", mode))))):
            raise ValueError("Color_range does not match mode")

        self.mode = mode

        if isinstance(fill_value, (tuple, list, set)):
            self.fill_value = list(fill_value)
        elif fill_value is not None:
            self.fill_value = [fill_value]
        else:
            self.fill_value = None

        self.channels = []
        self.palette = palette

        if isinstance(channels, (tuple, list)):
            if _areinstances(channels, (np.ma.core.MaskedArray, np.ndarray,
                                        list, tuple)):
                for i, chn in enumerate(channels):
                    if color_range is not None:
                        color_min = color_range[i][0]
                        color_max = color_range[i][1]
                    else:
                        color_min = 0.0
                        color_max = 1.0

                    # Add data to image object as a channel
                    self._add_channel(chn, color_min, color_max)

                    self.shape = self.channels[-1].shape
                    if self.shape != self.channels[0].shape:
                        raise ValueError("Channels must have the same shape.")
                self.height = self.shape[0]
                self.width = self.shape[1]
            else:
                raise ValueError("Channels must all be arrays, lists or "
                                 "tuples.")
        elif channels is not None:
            self.height = channels.shape[0]
            self.width = channels.shape[1]
            self.shape = channels.shape

            if color_range is not None:
                color_min = color_range[0]
                color_max = color_range[1]
            else:
                color_min = 0.0
                color_max = 1.0

            # Add data to image object as a channel
            self._add_channel(channels, color_min, color_max)

        else:
            self.shape = (0, 0)
            self.width = 0
            self.height = 0

    def _add_channel(self, chn, color_min, color_max):
        """Adds a channel to the image object
        """

        if isinstance(chn, np.ma.core.MaskedArray):
            chn_data = chn.data
            chn_mask = chn.mask
        else:
            chn_data = np.array(chn)
            chn_mask = False
        scaled = ((chn_data - color_min) *
                  1.0 / (color_max - color_min))
        self.channels.append(np.ma.array(scaled, mask=chn_mask))

    def _finalize(self, dtype=np.uint8):
        """Finalize the image, that is put it in RGB mode, and set the channels
        in 8bit format ([0,255] range).
        """
        channels = []
        if self.mode == "P":
            self.convert("RGB")
        if self.mode == "PA":
            self.convert("RGBA")

        for chn in self.channels:
            if isinstance(chn, np.ma.core.MaskedArray):
                final_data = chn.data.clip(0, 1) * np.iinfo(dtype).max
            else:
                final_data = chn.clip(0, 1) * np.iinfo(dtype).max

            channels.append(np.ma.array(final_data,
                                        dtype,
                                        mask=chn.mask))
        if self.fill_value is not None:
            fill_value = [int(col * np.iinfo(dtype).max)
                          for col in self.fill_value]
        else:
            fill_value = None
        return channels, fill_value

    def is_empty(self):
        """Checks for an empty image.
        """
        if(((self.channels == []) and (not self.shape == (0, 0))) or
           ((not self.channels == []) and (self.shape == (0, 0)))):
            raise RuntimeError("Channels-shape mismatch.")
        return self.channels == [] and self.shape == (0, 0)

    def show(self):
        """Display the image on screen.
        """
        self.pil_image().show()

    def pil_image(self):
        """Return a PIL image from the current image.
        """
        channels, fill_value = self._finalize()

        if self.is_empty():
            return Pil.new(self.mode, (0, 0))

        if(self.mode == "L"):
            if fill_value is not None:
                img = Pil.fromarray(channels[0].filled(fill_value))
            else:
                img = Pil.fromarray(channels[0].filled(0))
                alpha = np.zeros(channels[0].shape, np.uint8)
                mask = np.ma.getmaskarray(channels[0])
                alpha = np.where(mask, alpha, 255)
                pil_alpha = Pil.fromarray(alpha)

                img = Pil.merge("LA", (img, pil_alpha))
        elif(self.mode == "LA"):
            if fill_value is not None:
                img = Pil.fromarray(channels[0].filled(fill_value))
                pil_alpha = Pil.fromarray(channels[1])
            else:
                img = Pil.fromarray(channels[0].filled(0))
                alpha = np.zeros(channels[0].shape, np.uint8)
                mask = np.ma.getmaskarray(channels[0])
                alpha = np.where(mask, alpha, channels[1])
                pil_alpha = Pil.fromarray(alpha)
            img = Pil.merge("LA", (img, pil_alpha))

        elif(self.mode == "RGB"):
            # Mask where all channels have missing data (incomplete data will
            # be shown).
            mask = (np.ma.getmaskarray(channels[0]) &
                    np.ma.getmaskarray(channels[1]) &
                    np.ma.getmaskarray(channels[2]))

            if fill_value is not None:
                pil_r = Pil.fromarray(channels[0].filled(fill_value[0]))
                pil_g = Pil.fromarray(channels[1].filled(fill_value[1]))
                pil_b = Pil.fromarray(channels[2].filled(fill_value[2]))
                img = Pil.merge("RGB", (pil_r, pil_g, pil_b))
            else:
                pil_r = Pil.fromarray(channels[0].filled(0))
                pil_g = Pil.fromarray(channels[1].filled(0))
                pil_b = Pil.fromarray(channels[2].filled(0))

                alpha = np.zeros(channels[0].shape, np.uint8)
                alpha = np.where(mask, alpha, 255)
                pil_a = Pil.fromarray(alpha)

                img = Pil.merge("RGBA", (pil_r, pil_g, pil_b, pil_a))

        elif(self.mode == "RGBA"):
            # Mask where all channels have missing data (incomplete data will
            # be shown).
            mask = (np.ma.getmaskarray(channels[0]) &
                    np.ma.getmaskarray(channels[1]) &
                    np.ma.getmaskarray(channels[2]) &
                    np.ma.getmaskarray(channels[3]))

            if fill_value is not None:
                pil_r = Pil.fromarray(channels[0].filled(fill_value[0]))
                pil_g = Pil.fromarray(channels[1].filled(fill_value[1]))
                pil_b = Pil.fromarray(channels[2].filled(fill_value[2]))
                pil_a = Pil.fromarray(channels[3].filled(fill_value[3]))
                img = Pil.merge("RGBA", (pil_r, pil_g, pil_b, pil_a))
            else:
                pil_r = Pil.fromarray(channels[0].filled(0))
                pil_g = Pil.fromarray(channels[1].filled(0))
                pil_b = Pil.fromarray(channels[2].filled(0))

                alpha = np.where(mask, 0, channels[3])
                pil_a = Pil.fromarray(alpha)

                img = Pil.merge("RGBA", (pil_r, pil_g, pil_b, pil_a))

        else:
            raise TypeError("Does not know how to use mode %s." % (self.mode))

        return img

    def save(self, filename, compression=6, fformat=None):
        """Save the image to the given *filename*. For some formats like jpg
        and png, the work is delegated to :meth:`pil_save`, which doesn't
        support the *compression* option.
        """
        self.pil_save(filename, compression, fformat)

    def pil_save(self, filename, compression=6, fformat=None):
        """Save the image to the given *filename* using PIL. For now, the
        compression level [0-9] is ignored, due to PIL's lack of support. See
        also :meth:`save`.
        """
        # PIL does not support compression option.
        del compression

        if self.is_empty():
            raise IOError("Cannot save an empty image")

        ensure_dir(filename)

        fformat = fformat or os.path.splitext(filename)[1][1:4]
        fformat = check_image_format(fformat)

        params = {}

        if fformat == 'png':
            # Take care of GeoImage.tags.
            params['pnginfo'] = self._pngmeta()

        self.pil_image().save(filename, fformat, **params)

    def _pngmeta(self):
        """It will return GeoImage.tags as a PNG metadata object.

         Inspired by:
         public domain, Nick Galbreath
         http://blog.modp.com/2007/08/python-pil-and-png-metadata-take-2.html
        """
        reserved = ('interlace', 'gamma', 'dpi', 'transparency', 'aspect')

        try:
            tags = self.tags
        except AttributeError:
            tags = {}

        # Undocumented class
        from PIL import PngImagePlugin
        meta = PngImagePlugin.PngInfo()

        # Copy from tags to new dict
        for k__, v__ in tags.items():
            if k__ not in reserved:
                meta.add_text(k__, v__, 0)

        return meta

    def putalpha(self, alpha):
        """Adds an *alpha* channel to the current image, or replaces it with
        *alpha* if it already exists.
        """
        alpha = np.ma.array(alpha)
        if(not (alpha.shape[0] == 0 and
                self.shape[0] == 0) and
           alpha.shape != self.shape):
            raise ValueError("Alpha channel shape should match image shape")

        if(not self.mode.endswith("A")):
            self.convert(self.mode + "A")
        if not self.is_empty():
            self.channels[-1] = alpha

    def _rgb2ycbcr(self, mode):
        """Convert the image from RGB mode to YCbCr."""

        self._check_modes(("RGB", "RGBA"))

        (self.channels[0], self.channels[1], self.channels[2]) = \
            rgb2ycbcr(self.channels[0],
                      self.channels[1],
                      self.channels[2])

        if self.fill_value is not None:
            self.fill_value[0:3] = rgb2ycbcr(self.fill_value[0],
                                             self.fill_value[1],
                                             self.fill_value[2])

        self.mode = mode

    def _ycbcr2rgb(self, mode):
        """Convert the image from YCbCr mode to RGB.
        """

        self._check_modes(("YCbCr", "YCbCrA"))

        (self.channels[0], self.channels[1], self.channels[2]) = \
            ycbcr2rgb(self.channels[0],
                      self.channels[1],
                      self.channels[2])

        if self.fill_value is not None:
            self.fill_value[0:3] = ycbcr2rgb(self.fill_value[0],
                                             self.fill_value[1],
                                             self.fill_value[2])

        self.mode = mode

    def _to_p(self, mode):
        """Convert the image to P or PA mode.
        """

        if self.mode.endswith("A"):
            chans = self.channels[:-1]
            alpha = self.channels[-1]
            self._secondary_mode = self.mode[:-1]
        else:
            chans = self.channels
            alpha = None
            self._secondary_mode = self.mode

        palette = []
        selfmask = reduce(np.ma.mask_or, [chn.mask for chn in chans])
        new_chn = np.ma.zeros(self.shape, dtype=int)
        color_nb = 0

        for i in range(self.height):
            for j in range(self.width):
                current_col = tuple([chn[i, j] for chn in chans])
                try:
                    (idx
                     for idx in range(len(palette))
                     if palette[idx] == current_col).next()
                except StopIteration:
                    idx = color_nb
                    palette.append(current_col)
                    color_nb = color_nb + 1

                new_chn[i, j] = idx

        if self.fill_value is not None:
            if self.mode.endswith("A"):
                current_col = tuple(self.fill_value[:-1])
                fill_alpha = [self.fill_value[-1]]
            else:
                current_col = tuple(self.fill_value)
                fill_alpha = []
            try:
                (idx
                 for idx in range(len(palette))
                 if palette[idx] == current_col).next()
            except StopIteration:
                idx = color_nb
                palette.append(current_col)
                color_nb = color_nb + 1

            self.fill_value = [idx] + fill_alpha

        new_chn.mask = selfmask
        self.palette = palette
        if alpha is None:
            self.channels = [new_chn]
        else:
            self.channels = [new_chn, alpha]
        self.mode = mode

    def _from_p(self, mode):
        """Convert the image from P or PA mode.
        """

        self._check_modes(("P", "PA"))

        if self.mode.endswith("A"):
            alpha = self.channels[-1]
        else:
            alpha = None

        chans = []
        cdfs = []

        color_chan = self.channels[0]

        for i in range(len(self.palette[0])):
            cdfs.append(np.zeros(len(self.palette)))
            for j in range(len(self.palette)):
                cdfs[i][j] = self.palette[j][i]
            new_chn = np.ma.array(np.interp(color_chan,
                                            np.arange(len(self.palette)),
                                            cdfs[i]),
                                  mask=color_chan.mask)
            chans.append(new_chn)

        if self.fill_value is not None:
            if alpha is not None:
                fill_alpha = self.fill_value[-1]
                self.fill_value = list(self.palette[int(self.fill_value[0])])
                self.fill_value += [fill_alpha]
            else:
                self.fill_value = list(self.palette[int(self.fill_value[0])])

        self.mode = self._secondary_mode
        self.channels = chans
        if alpha is not None:
            self.channels.append(alpha)
            self.mode = self.mode + "A"

        self.convert(mode)

    def _check_modes(self, modes):
        """Check that the image is in on of the given *modes*, raise an
        exception otherwise.
        """
        if not isinstance(modes, (tuple, list, set)):
            modes = [modes]
        if self.mode not in modes:
            raise ValueError("Image not in suitable mode: %s" % modes)

    def _l2rgb(self, mode):
        """Convert from L (black and white) to RGB.
        """
        self._check_modes(("L", "LA"))
        self.channels.append(self.channels[0].copy())
        self.channels.append(self.channels[0].copy())
        if self.fill_value is not None:
            self.fill_value = self.fill_value[:1] * 3 + self.fill_value[1:]
        if self.mode == "LA":
            self.channels[1], self.channels[3] = \
                self.channels[3], self.channels[1]
        self.mode = mode

    def _rgb2l(self, mode):
        """Convert from RGB to monochrome L.
        """
        self._check_modes(("RGB", "RGBA"))

        kb_ = 0.114
        kr_ = 0.299

        r__ = self.channels[0]
        g__ = self.channels[1]
        b__ = self.channels[2]

        y__ = kr_ * r__ + (1 - kr_ - kb_) * g__ + kb_ * b__

        if self.fill_value is not None:
            self.fill_value = ([rgb2ycbcr(self.fill_value[0],
                                          self.fill_value[1],
                                          self.fill_value[2])[0]] +
                               self.fill_value[3:])

        self.channels = [y__] + self.channels[3:]

        self.mode = mode

    def _ycbcr2l(self, mode):
        """Convert from YCbCr to L.
        """
        self._check_modes(("YCbCr", "YCbCrA"))

        self.channels = [self.channels[0]] + self.channels[3:]
        if self.fill_value is not None:
            self.fill_value = [self.fill_value[0]] + self.fill_value[3:]
        self.mode = mode

    def _l2ycbcr(self, mode):
        """Convert from L to YCbCr.
        """
        self._check_modes(("L", "LA"))

        luma = self.channels[0]
        zeros = np.ma.zeros(luma.shape)
        zeros.mask = luma.mask

        self.channels = [luma, zeros, zeros] + self.channels[1:]

        if self.fill_value is not None:
            self.fill_value = [self.fill_value[0], 0, 0] + self.fill_value[1:]

        self.mode = mode

    def convert(self, mode):
        """Convert the current image to the given *mode*. See :class:`Image`
        for a list of available modes.
        """
        if mode == self.mode:
            return

        if mode not in ["L", "LA", "RGB", "RGBA",
                        "YCbCr", "YCbCrA", "P", "PA"]:
            raise ValueError("Mode %s not recognized." % (mode))

        if self.is_empty():
            self.mode = mode
            return

        if(mode == self.mode + "A"):
            self.channels.append(np.ma.ones(self.channels[0].shape))
            if self.fill_value is not None:
                self.fill_value += [1]
            self.mode = mode

        elif(mode + "A" == self.mode):
            self.channels = self.channels[:-1]
            if self.fill_value is not None:
                self.fill_value = self.fill_value[:-1]
            self.mode = mode

        elif(mode.endswith("A") and not self.mode.endswith("A")):
            self.convert(self.mode + "A")
            self.convert(mode)

        elif(self.mode.endswith("A") and not mode.endswith("A")):
            self.convert(self.mode[:-1])
            self.convert(mode)

        else:
            cases = {
                "RGB": {"YCbCr": self._rgb2ycbcr,
                        "L": self._rgb2l,
                        "P": self._to_p},
                "RGBA": {"YCbCrA": self._rgb2ycbcr,
                         "LA": self._rgb2l,
                         "PA": self._to_p},
                "YCbCr": {"RGB": self._ycbcr2rgb,
                          "L": self._ycbcr2l,
                          "P": self._to_p},
                "YCbCrA": {"RGBA": self._ycbcr2rgb,
                           "LA": self._ycbcr2l,
                           "PA": self._to_p},
                "L": {"RGB": self._l2rgb,
                      "YCbCr": self._l2ycbcr,
                      "P": self._to_p},
                "LA": {"RGBA": self._l2rgb,
                       "YCbCrA": self._l2ycbcr,
                       "PA": self._to_p},
                "P": {"RGB": self._from_p,
                      "YCbCr": self._from_p,
                      "L": self._from_p},
                "PA": {"RGBA": self._from_p,
                       "YCbCrA": self._from_p,
                       "LA": self._from_p}}
            try:
                cases[self.mode][mode](mode)
            except KeyError:
                raise ValueError("Conversion from %s to %s not implemented !"
                                 % (self.mode, mode))

    def clip(self, channels=True):
        """Limit the values of the array to the default [0,1] range. *channels*
        says which channels should be clipped."""
        if not (isinstance(channels, (tuple, list))):
            channels = [channels] * len(self.channels)

        for i in range(len(self.channels)):
            if channels[i]:
                self.channels[i] = np.ma.clip(self.channels[i], 0.0, 1.0)

    def resize(self, shape):
        """Resize the image to the given *shape* tuple, in place. For zooming,
        nearest neighbour method is used, while for shrinking, decimation is
        used. Therefore, *shape* must be a multiple or a divisor of the image
        shape.
        """
        if self.is_empty():
            raise ValueError("Cannot resize an empty image")

        factor = [1, 1]
        zoom = [True, True]
        zoom[0] = shape[0] >= self.height
        zoom[1] = shape[1] >= self.width

        if zoom[0]:
            factor[0] = shape[0] * 1.0 / self.height
        else:
            factor[0] = self.height * 1.0 / shape[0]
        if zoom[1]:
            factor[1] = shape[1] * 1.0 / self.width
        else:
            factor[1] = self.width * 1.0 / shape[1]

        if(int(factor[0]) != factor[0] or
           int(factor[1]) != factor[1]):
            raise ValueError("Resize not of integer factor!")

        factor[0] = int(factor[0])
        factor[1] = int(factor[1])

        i = 0
        for chn in self.channels:
            if zoom[0]:
                chn = chn.repeat([factor[0]] * chn.shape[0], axis=0)
            else:
                chn = chn[[idx * factor[0]
                           for idx in range(self.height / factor[0])],
                          :]
            if zoom[1]:
                self.channels[i] = chn.repeat([factor[1]] * chn.shape[1],
                                              axis=1)
            else:
                self.channels[i] = chn[:,
                                       [idx * factor[1]
                                        for idx in range(self.width /
                                                         factor[1])]]

            i = i + 1

        self.height = self.channels[0].shape[0]
        self.width = self.channels[0].shape[1]
        self.shape = self.channels[0].shape

    def replace_luminance(self, luminance):
        """Replace the Y channel of the image by the array *luminance*. If the
        image is not in YCbCr mode, it is converted automatically to and
        from that mode.
        """
        if self.is_empty():
            return

        if (luminance.shape != self.channels[0].shape):
            if ((luminance.shape[0] * 1.0 / luminance.shape[1]) ==
                    (self.channels[0].shape[0] * 1.0 / self.channels[0].shape[1])):
                if luminance.shape[0] > self.channels[0].shape[0]:
                    self.resize(luminance.shape)
                else:
                    raise NameError("Luminance smaller than the image !")
            else:
                raise NameError("Not the good shape !")

        mode = self.mode
        if mode.endswith("A"):
            self.convert("YCbCrA")
            self.channels[0] = luminance
            self.convert(mode)
        else:
            self.convert("YCbCr")
            self.channels[0] = luminance
            self.convert(mode)

    def enhance(self, inverse=False, gamma=1.0, stretch="no"):
        """Image enhancement function. It applies **in this order** inversion,
        gamma correction, and stretching to the current image, with parameters
        *inverse* (see :meth:`Image.invert`), *gamma* (see
        :meth:`Image.gamma`), and *stretch* (see :meth:`Image.stretch`).
        """
        self.invert(inverse)
        self.gamma(gamma)
        self.stretch(stretch)

    def gamma(self, gamma=1.0):
        """Apply gamma correction to the channels of the image. If *gamma* is a
        tuple, then it should have as many elements as the channels of the
        image, and the gamma correction is applied elementwise. If *gamma* is a
        number, the same gamma correction is applied on every channel, if there
        are several channels in the image. The behaviour of :func:`gamma` is
        undefined outside the normal [0,1] range of the channels.
        """
        if not isinstance(gamma, (int, long, float)):
            if(not isinstance(gamma, (tuple, list, set)) or
               not _areinstances(gamma, (int, long, float))):
                raise TypeError("Gamma should be a real number, or an iterable "
                                "of real numbers.")

        if(isinstance(gamma, (list, tuple, set)) and
           len(gamma) != len(self.channels)):
            raise ValueError("Number of channels and gamma components differ.")

        if gamma < 0:
            raise ValueError("Gamma correction must be a positive number.")

        if gamma == 1.0:
            return
        if (isinstance(gamma, (tuple, list))):
            gamma_list = list(gamma)
        else:
            gamma_list = [gamma] * len(self.channels)
        for i in range(len(self.channels)):
            if(isinstance(self.channels[i], np.ma.core.MaskedArray)):
                if ne:
                    self.channels[i] = np.ma.array(
                        ne.evaluate("data ** (1.0 / gamma)",
                                    local_dict={"data": self.channels[i].data,
                                                'gamma': gamma_list[i]}),
                        mask=self.channels[i].mask,
                        copy=False)
                else:
                    self.channels[i] = np.ma.array(self.channels[i].data **
                                                   (1.0 / gamma_list[i]),
                                                   mask=self.channels[i].mask,
                                                   copy=False)
            else:
                self.channels[i] = np.where(self.channels[i] >= 0,
                                            self.channels[i] **
                                            (1.0 / gamma_list[i]),
                                            self.channels[i])

    def stretch(self, stretch="no", **kwarg):
        """Apply stretching to the current image. The value of *stretch* sets
        the type of stretching applied. The values "histogram", "linear",
        "crude" (or "crude-stretch") perform respectively histogram
        equalization, contrast stretching (with 5% cutoff on both sides), and
        contrast stretching without cutoff. The value "logarithmic" or "log"
        will do a logarithmic enhancement towards white. If a tuple or a list
        of two values is given as input, then a contrast stretching is performed
        with the values as cutoff. These values should be normalized in the
        range [0.0,1.0].
        """
        if((isinstance(stretch, tuple) or
                isinstance(stretch, list))):
            if len(stretch) == 2:
                for i in range(len(self.channels)):
                    self.stretch_linear(i, cutoffs=stretch, **kwarg)
            else:
                raise ValueError(
                    "Stretch tuple must have exactly two elements")
        elif stretch == "linear":
            for i in range(len(self.channels)):
                self.stretch_linear(i, **kwarg)
        elif stretch == "histogram":
            for i in range(len(self.channels)):
                self.stretch_hist_equalize(i, **kwarg)
        elif(stretch in ["crude", "crude-stretch"]):
            for i in range(len(self.channels)):
                self.crude_stretch(i, **kwarg)
        elif(stretch in ["log", "logarithmic"]):
            for i in range(len(self.channels)):
                self.stretch_logarithmic(i, **kwarg)
        elif(stretch == "no"):
            return
        elif isinstance(stretch, str):
            raise ValueError("Stretching method %s not recognized." % stretch)
        else:
            raise TypeError("Stretch parameter must be a string or a tuple.")

    def invert(self, invert=True):
        """Inverts all the channels of a image according to *invert*. If invert
        is a tuple or a list, elementwise invertion is performed, otherwise all
        channels are inverted if *invert* is true (default).
        """
        if(isinstance(invert, (tuple, list, set)) and
           len(self.channels) != len(invert)):
            raise ValueError(
                "Number of channels and invert components differ.")

        if isinstance(invert, (tuple, list, set)):
            i = 0
            for chn in self.channels:
                if(invert[i]):
                    self.channels[i] = 1.0 - chn
                i = i + 1
        elif(invert):
            i = 0
            for chn in self.channels:
                self.channels[i] = 1.0 - chn
                i = i + 1

    def stretch_hist_equalize(self, ch_nb):
        """Stretch the current image's colors by performing histogram
        equalization on channel *ch_nb*.
        """
        LOG.info("Perform a histogram equalized contrast stretch.")

        if(self.channels[ch_nb].size ==
           np.ma.count_masked(self.channels[ch_nb])):
            LOG.warning("Nothing to stretch !")
            return

        arr = self.channels[ch_nb]

        nwidth = 2048.0

        carr = arr.compressed()

        cdf = np.arange(0.0, 1.0, 1 / nwidth)
        LOG.debug("Make histogram bins having equal amount of data, " +
                  "using numpy percentile function:")
        bins = percentile(carr, list(cdf * 100))

        res = np.ma.empty_like(arr)
        res.mask = np.ma.getmaskarray(arr)
        res[~res.mask] = np.interp(carr, bins, cdf)

        self.channels[ch_nb] = res

    def stretch_logarithmic(self, ch_nb, factor=100.):
        """Move data into range [1:factor] and do a normalized logarithmic
        enhancement.
        """
        LOG.debug("Perform a logarithmic contrast stretch.")
        if ((self.channels[ch_nb].size ==
             np.ma.count_masked(self.channels[ch_nb])) or
                (self.channels[ch_nb].min() == self.channels[ch_nb].max())):
            LOG.warning("Nothing to stretch !")
            return

        crange = (0., 1.0)

        arr = self.channels[ch_nb]
        b = float(crange[1] - crange[0]) / np.log(factor)
        c = float(crange[0])
        slope = (factor - 1.) / float(arr.max() - arr.min())
        arr = 1. + (arr - arr.min()) * slope
        arr = c + b * np.log(arr)
        self.channels[ch_nb] = arr

    def stretch_linear(self, ch_nb, cutoffs=(0.005, 0.005)):
        """Stretch linearly the contrast of the current image on channel
        *ch_nb*, using *cutoffs* for left and right trimming.
        """
        LOG.debug("Perform a linear contrast stretch.")

        if((self.channels[ch_nb].size ==
            np.ma.count_masked(self.channels[ch_nb])) or
           self.channels[ch_nb].min() == self.channels[ch_nb].max()):
            LOG.warning("Nothing to stretch !")
            return

        arr = self.channels[ch_nb]
        carr = arr.compressed()

        LOG.debug("Calculate the histogram percentiles: ")
        LOG.debug("Left and right percentiles: " +
                  str(cutoffs[0] * 100) + " " + str(cutoffs[1] * 100))

        left, right = percentile(carr, [cutoffs[0] * 100,
                                        100. - cutoffs[1] * 100])

        delta_x = (right - left)
        LOG.debug("Interval: left=%f, right=%f width=%f"
                  % (left, right, delta_x))

        if delta_x > 0.0:
            self.channels[ch_nb] = np.ma.array((arr - left) / delta_x,
                                               mask=arr.mask)
        else:
            self.channels[ch_nb] = np.ma.zeros(arr.shape)
            LOG.warning("Unable to make a contrast stretch!")

    def crude_stretch(self, ch_nb, min_stretch=None, max_stretch=None):
        """Perform simple linear stretching (without any cutoff) on the channel
        *ch_nb* of the current image and normalize to the [0,1] range."""

        if(min_stretch is None):
            min_stretch = self.channels[ch_nb].min()
        if(max_stretch is None):
            max_stretch = self.channels[ch_nb].max()

        if((not self.channels[ch_nb].mask.all()) and
                max_stretch - min_stretch > 0):
            stretched = self.channels[ch_nb].data.astype(np.float)
            stretched -= min_stretch
            stretched /= max_stretch - min_stretch
            self.channels[ch_nb] = np.ma.array(stretched,
                                               mask=self.channels[ch_nb].mask,
                                               copy=False)
        else:
            LOG.warning("Nothing to stretch !")

    def merge(self, img):
        """Use the provided image as background for the current *img* image,
        that is if the current image has missing data.
        """
        if self.is_empty():
            raise ValueError("Cannot merge an empty image.")

        if(self.mode != img.mode):
            raise ValueError("Cannot merge image of different modes.")

        selfmask = reduce(np.ma.mask_or, [chn.mask for chn in self.channels])

        for i in range(len(self.channels)):
            self.channels[i] = np.ma.where(selfmask,
                                           img.channels[i],
                                           self.channels[i])
            self.channels[i].mask = np.logical_and(selfmask,
                                                   img.channels[i].mask)


def all(iterable):
    for element in iterable:
        if not element:
            return False
    return True


def _areinstances(the_list, types):
    """Check if all the elements of the list are of given type.
    """
    return all([isinstance(item, types) for item in the_list])


def _is_pair(item):
    """Check if an item is a pair (tuple of size 2).
    """
    return (isinstance(item, (list, tuple, set)) and
            len(item) == 2 and
            not isinstance(item[0], (list, tuple, set)) and
            not isinstance(item[1], (list, tuple, set)))


def _is_list_of_pairs(the_list):
    """Check if a list contains only pairs.
    """
    return all([_is_pair(item) for item in the_list])


def ycbcr2rgb(y__, cb_, cr_):
    """Convert the three YCbCr channels to RGB channels.
    """

    kb_ = 0.114
    kr_ = 0.299

    r__ = 2 * cr_ / (1 - kr_) + y__
    b__ = 2 * cb_ / (1 - kb_) + y__
    g__ = (y__ - kr_ * r__ - kb_ * b__) / (1 - kr_ - kb_)

    return r__, g__, b__


def rgb2ycbcr(r__, g__, b__):
    """Convert the three RGB channels to YCbCr."""

    kb_ = 0.114
    kr_ = 0.299

    y__ = kr_ * r__ + (1 - kr_ - kb_) * g__ + kb_ * b__
    cb_ = 1. / (2 * (1 - kb_)) * (b__ - y__)
    cr_ = 1. / (2 * (1 - kr_)) * (r__ - y__)

    return y__, cb_, cr_



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2012, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to PPS level1 radiance data in ODIM format.
"""
import glob
import os.path
from ConfigParser import ConfigParser

from datetime import datetime, timedelta
import math
import numpy as np

from mpop import CONFIG_PATH
import logging

LOG = logging.getLogger(__name__)

import h5py


EPSILON = 0.001


class InfoObject(object):

    """Simple data and info container.
    """

    def __init__(self):
        self.info = {}
        self.data = None


def pack_signed(data, data_type):
    bits = np.iinfo(data_type).bits
    scale_factor = (data.max() - data.min()) / (2 ** bits - 2)
    add_offset = (data.max() - data.min()) / 2
    no_data = - 2 ** (bits - 1)
    pack = ((data - add_offset) / scale_factor).astype(data_type)
    return pack, scale_factor, add_offset, no_data


def load(satscene, *args, **kwargs):
    """Read data from file and load it into *satscene*.
    """
    del args, kwargs
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level2",
                                    raw=True):
        options[option] = value
    CASES[satscene.instrument_name](satscene, options)


def load_channels(satscene, options):
    """Read avhrr/viirs/modis radiance (tb's and refl) data from file and load
    it into *satscene*.
    """

    if "filename" not in options:
        raise IOError("No filename given, cannot load.")

    LOG.debug("Start loading channels")

    if satscene.instrument_name in ['avhrr']:
        chns = satscene.channels_to_load & set(
            ["1", "2", "3A", "3B", "4", "5"])
    elif satscene.instrument_name in ['viirs']:
        chns = satscene.channels_to_load & set(["M05", "M07", "M10",
                                                "M12", "M14", "M15", "M16"])

    if len(chns) == 0:
        return

    values = {"orbit": satscene.orbit,
              "satname": satscene.satname,
              "number": satscene.number,
              "instrument": satscene.instrument_name,
              "satellite": satscene.fullname
              }

    filename = os.path.join(satscene.time_slot.strftime(options["dir"]) % values,
                            satscene.time_slot.strftime(options["filename"])
                            % values)

    file_list = glob.glob(filename)

    if len(file_list) > 1:
        raise IOError("More than one l1b file matching!")
    elif len(file_list) == 0:
        raise IOError("No PPS level 1 file matching!: " +
                      filename)

    filename = file_list[0]

    LOG.debug("Loading from " + filename)

    available_channels = set([])
    data_channels = {}
    info_channels = {}
    instrument_data = NwcSafPpsOdim(filename)

    idx = 1
    while hasattr(instrument_data, 'image%d' % idx):
        channel = getattr(instrument_data, 'image%d' % idx)
        channel_name = channel.info['channel'].upper()
        #channel_des = channel.info['description']
        available_channels |= set([channel_name])
        data_channels[channel_name] = channel.data
        info_channels[channel_name] = channel.info
        idx = idx + 1

    for chn in satscene.channels_to_load:
        if chn in available_channels:
            if info_channels[chn]['quantity'] in ["REFL"]:
                units = "%"
            elif info_channels[chn]['quantity'] in ["TB"]:
                units = "K"
            else:
                LOG.warning("Units not known! Unit = " +
                            str(info_channels[chn]['quantity']))

            gain = info_channels[chn]["gain"]
            intercept = info_channels[chn]["offset"]

            chn_array = np.ma.array(data_channels[chn])
            missing_data = info_channels[chn]["missingdata"]
            chn_array = np.ma.masked_inside(chn_array,
                                            missing_data - EPSILON,
                                            missing_data + EPSILON)
            no_data = info_channels[chn]["nodata"]
            chn_array = np.ma.masked_inside(chn_array,
                                            no_data - EPSILON,
                                            no_data + EPSILON)

            satscene[chn] = chn_array
            satscene[chn].data = np.ma.masked_less(satscene[chn].data *
                                                   gain +
                                                   intercept,
                                                   0)

            satscene[chn].info['units'] = units
        else:
            LOG.warning("Channel " + str(chn) + " not available, not loaded.")

    # Compulsory global attributes
    satscene.info["title"] = (satscene.satname.capitalize() + satscene.number +
                              " satellite, " +
                              satscene.instrument_name.capitalize() +
                              " instrument.")
    satscene.info["institution"] = "Data processed by EUMETSAT NWCSAF/PPS."
    satscene.add_to_history("PPS level 1 data read by mpop.")
    satscene.info["references"] = "No reference."
    satscene.info["comments"] = "No comment."

    lons = (instrument_data.lon.data * instrument_data.lon.info['gain'] +
            instrument_data.lon.info['offset'])
    lats = (instrument_data.lat.data * instrument_data.lat.info['gain'] +
            instrument_data.lat.info['offset'])

    try:
        from pyresample import geometry
        satscene.area = geometry.SwathDefinition(lons=lons, lats=lats)
    except ImportError:
        satscene.area = None
        satscene.lat = lats
        satscene.lon = lons


class NwcSafPpsOdim(object):

    def __init__(self, filename=None):
        self._how = {}
        self._what = {}
        self._projectables = []
        self._keys = []
        self._refs = {}
        self.lon = None
        self.lat = None
        self.shape = None
        if filename:
            self.read(filename)

    def read(self, filename):
        """Read data in hdf5 format from *filename*
        """

        h5f = h5py.File(filename, "r")

        # Read the /how attributes

        self._how = dict(h5f['how'].attrs)
        self._what = dict(h5f['what'].attrs)
        self._how["satellite"] = h5f['how'].attrs['platform']
        # Which one to use?:
        self._how["time_slot"] = (timedelta(seconds=long(h5f['how'].attrs['startepochs']))
                                  + datetime(1970, 1, 1, 0, 0))
        self._what["time_slot"] = datetime.strptime(h5f['what'].attrs['date'] +
                                                    h5f['what'].attrs['time'],
                                                    "%Y%m%d%H%M%S")

        # Read the data and attributes
        #   This covers only one level of data. This could be made recursive.
        for key, dataset in h5f.iteritems():
            if "how" in dataset.name or "what" in dataset.name:
                continue

            if "image" in dataset.name:
                setattr(self, key, InfoObject())
                getattr(self, key).info = dict(dataset.attrs)
                getattr(self, key).data = dataset['data'][:]

                if 'how' in dataset:
                    for skey, value in dataset['how'].attrs.iteritems():
                        getattr(self, key).info[skey] = value
                if 'what' in dataset:
                    for skey, value in dataset['what'].attrs.iteritems():
                        getattr(self, key).info[skey] = value

            if "where" in dataset.name:
                setattr(self, 'lon', InfoObject())
                getattr(self, 'lon').data = h5f['/where/lon/data'][:]
                getattr(self, 'lon').info = dict(dataset.attrs)
                for skey, value in dataset['lon/what'].attrs.iteritems():
                    getattr(self, 'lon').info[skey] = value

                setattr(self, 'lat', InfoObject())
                getattr(self, 'lat').data = h5f['/where/lat/data'][:]
                getattr(self, 'lat').info = dict(dataset.attrs)
                for skey, value in dataset['lat/what'].attrs.iteritems():
                    getattr(self, 'lat').info[skey] = value

                self.shape = self.lon.data.shape

        h5f.close()

        # Setup geolocation

        try:
            from pyresample import geometry
        except ImportError:
            return

        if hasattr(self, "lon") and hasattr(self, "lat"):
            lons = self.lon.data * \
                self.lon.info["gain"] + self.lon.info["gain"]
            lats = self.lat.data * \
                self.lat.info["gain"] + self.lat.info["gain"]
            self.area = geometry.SwathDefinition(lons=lons, lats=lats)
        else:
            LOG.warning("No longitudes or latitudes for data")

    # def project(self, coverage):
    #     """Project what can be projected in the product.
    #     """
    #     import copy
    #     res = copy.copy(self)

    #     area = coverage.out_area

    # Project the data
    #     for var in self._projectables:
    #         LOG.info("Projecting " + str(var))
    #         res.__dict__[var] = copy.copy(self.__dict__[var])
    #         res.__dict__[var].data = coverage.project_array(
    #             self.__dict__[var].data)

    # Take care of geolocation
    # We support only swath data with full lon,lat arrays
    #     lons, scale_factor, add_offset, no_data = \
    #         pack_signed(area.lons[:], np.int16)
    #     res.lon = InfoObject()
    #     res.lon.data = lons
    #     res.lon.info["description"] = "geographic longitude (deg)"
    #     res.lon.info["offset"] = add_offset
    #     res.lon.info["gain"] = scale_factor
    #     res.lon.info["nodata"] = no_data
    #     if "lon" not in res._keys:
    #         res._keys.append("lon")

    #     lats, scale_factor, add_offset, no_data = \
    #         pack_signed(area.lats[:], np.int16)
    #     res.lat = InfoObject()
    #     res.lat.data = lats
    #     res.lat.info["description"] = "geographic latitude (deg)"
    #     res.lat.info["offset"] = add_offset
    #     res.lat.info["gain"] = scale_factor
    #     res.lat.info["nodata"] = no_data
    #     if "lat" not in res._keys:
    #         res._keys.append("lat")

    #     return res

    # def is_loaded(self):
    #     """Tells if the channel contains loaded data.
    #     """
    #     return len(self._projectables) > 0

CASES = {
    "avhrr": load_channels,
    "viirs": load_channels
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014, 2015 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Compact viirs format.
"""

import h5py
import numpy as np
from pyresample.geometry import SwathDefinition
from datetime import timedelta
import glob
from ConfigParser import ConfigParser
import os
import logging
import bz2

try:
    import tables
except ImportError:
    tables = None

from mpop import CONFIG_PATH

logger = logging.getLogger(__name__)

c = 299792458  # m.s-1
h = 6.6260755e-34  # m2kg.s-1
k = 1.380658e-23  # m2kg.s-2.K-1


def load(satscene, *args, **kwargs):
    del args

    files_to_load = []
    files_to_delete = []

    try:
        filename = kwargs.get("filename")
        logger.debug("reading %s", str(filename))
        if filename is not None:
            if isinstance(filename, (list, set, tuple)):
                files = filename
            else:
                files = [filename]
            files_to_load = []
            for filename in files:
                pathname, ext = os.path.splitext(filename)
                if ext == ".bz2":
                    zipfile = bz2.BZ2File(filename)
                    newname = os.path.join("/tmp", os.path.basename(pathname))
                    if not os.path.exists(newname):
                        with open(newname, "wb") as fp_:
                            fp_.write(zipfile.read())
                    zipfile.close()
                    files_to_load.append(newname)
                    files_to_delete.append(newname)
                else:
                    files_to_load.append(filename)
        else:
            time_start, time_end = kwargs.get("time_interval",
                                              (satscene.time_slot, None))

            conf = ConfigParser()
            conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
            options = {}
            for option, value in conf.items(satscene.instrument_name + "-level2",
                                            raw=True):
                options[option] = value

            template = os.path.join(options["dir"], options["filename"])

            second = timedelta(seconds=1)
            files_to_load = []

            if time_end is not None:
                time = time_start - second * 85
                files_to_load = []
                while time <= time_end:
                    fname = time.strftime(template)
                    flist = glob.glob(fname)
                    try:
                        files_to_load.append(flist[0])
                        time += second * 80
                    except IndexError:
                        pass
                    time += second

            else:
                files_to_load = glob.glob(time_start.strftime(template))

        chan_dict = {"M01": "M1",
                     "M02": "M2",
                     "M03": "M3",
                     "M04": "M4",
                     "M05": "M5",
                     "M06": "M6",
                     "M07": "M7",
                     "M08": "M8",
                     "M09": "M9",
                     "M10": "M10",
                     "M11": "M11",
                     "M12": "M12",
                     "M13": "M13",
                     "M14": "M14",
                     "M15": "M15",
                     "M16": "M16",
                     "DNB": "DNB"}

        channels = [(chn, chan_dict[chn])
                    for chn in satscene.channels_to_load
                    if chn in chan_dict]
        try:
            channels_to_load, chans = zip(*channels)
        except ValueError:
            return

        m_chans = []
        dnb_chan = []
        for chn in chans:
            if chn.startswith('M'):
                m_chans.append(chn)
            elif chn.startswith('DNB'):
                dnb_chan.append(chn)
            else:
                raise ValueError("Reading of channel %s not implemented", chn)

        m_datas = []
        m_lonlats = []
        dnb_datas = []
        dnb_lonlats = []

        for fname in files_to_load:
            is_dnb = os.path.basename(fname).startswith('SVDNBC')
            logger.debug("Reading %s", fname)
            if is_dnb:
                if tables:
                    h5f = tables.open_file(fname, "r")
                else:
                    logger.warning("DNB data could not be read from %s, "
                                   "PyTables not available.", fname)
                    continue
            else:
                h5f = h5py.File(fname, "r")
            if m_chans and not is_dnb:
                try:
                    arr, m_units = read_m(h5f, m_chans)
                    m_datas.append(arr)
                    m_lonlats.append(navigate_m(h5f, m_chans[0]))
                except KeyError:
                    pass
            if dnb_chan and is_dnb and tables:
                try:
                    arr, dnb_units = read_dnb(h5f)
                    dnb_datas.append(arr)
                    dnb_lonlats.append(navigate_dnb(h5f))
                except KeyError:
                    pass
            h5f.close()

        if len(m_lonlats) > 0:
            m_lons = np.ma.vstack([lonlat[0] for lonlat in m_lonlats])
            m_lats = np.ma.vstack([lonlat[1] for lonlat in m_lonlats])
        if len(dnb_lonlats) > 0:
            dnb_lons = np.ma.vstack([lonlat[0] for lonlat in dnb_lonlats])
            dnb_lats = np.ma.vstack([lonlat[1] for lonlat in dnb_lonlats])

        m_i = 0
        dnb_i = 0
        for chn in channels_to_load:
            if m_datas and chn.startswith('M'):
                m_data = np.ma.vstack([dat[m_i] for dat in m_datas])
                satscene[chn] = m_data
                satscene[chn].info["units"] = m_units[m_i]
                m_i += 1
            if dnb_datas and chn.startswith('DNB'):
                dnb_data = np.ma.vstack([dat[dnb_i] for dat in dnb_datas])
                satscene[chn] = dnb_data
                satscene[chn].info["units"] = dnb_units[dnb_i]
                dnb_i += 1

        if m_datas:
            m_area_def = SwathDefinition(np.ma.masked_where(m_data.mask, m_lons),
                                         np.ma.masked_where(m_data.mask, m_lats))
        else:
            logger.warning("No M channel data available.")

        if dnb_datas:
            dnb_area_def = SwathDefinition(np.ma.masked_where(dnb_data.mask,
                                                              dnb_lons),
                                           np.ma.masked_where(dnb_data.mask,
                                                              dnb_lats))
        else:
            logger.warning("No DNB data available.")

        for chn in channels_to_load:
            if "DNB" not in chn and m_datas:
                satscene[chn].area = m_area_def

        if dnb_datas:
            for chn in dnb_chan:
                satscene[chn].area = dnb_area_def

    finally:
        for fname in files_to_delete:
            if os.path.exists(fname):
                os.remove(fname)


def read_m(h5f, channels, calibrate=1):

    chan_dict = dict([(key.split("-")[1], key)
                      for key in h5f["All_Data"].keys()
                      if key.startswith("VIIRS")])

    scans = h5f["All_Data"]["NumberOfScans"][0]
    res = []
    units = []
    arr_mask = np.ma.nomask

    for channel in channels:
        rads = h5f["All_Data"][chan_dict[channel]]["Radiance"]
        if channel in ("M9",):
            arr = rads[:scans * 16, :].astype(np.float32)
            arr[arr > 65526] = np.nan
            arr = np.ma.masked_array(arr, mask=arr_mask)
        else:
            arr = np.ma.masked_greater(rads[:scans * 16, :].astype(np.float32),
                                       65526)
        try:
            arr = np.ma.where(arr <= rads.attrs['Threshold'],
                              arr * rads.attrs['RadianceScaleLow'] +
                              rads.attrs['RadianceOffsetLow'],
                              arr * rads.attrs['RadianceScaleHigh'] + \
                              rads.attrs['RadianceOffsetHigh'],)
            arr_mask = arr.mask
        except KeyError:
            print "KeyError"
            pass
        unit = "W m-2 sr-1 μm-1"
        if calibrate == 0:
            raise NotImplementedError("Can't get counts from this data")
        if calibrate == 1:
            # do calibrate
            try:
                # First guess: VIS or NIR data
                a_vis = rads.attrs['EquivalentWidth']
                b_vis = rads.attrs['IntegratedSolarIrradiance']
                dse = rads.attrs['EarthSunDistanceNormalised']
                arr *= 100 * np.pi * a_vis / b_vis * (dse ** 2)
                unit = "%"
            except KeyError:
                # Maybe it's IR data?
                try:
                    a_ir = rads.attrs['BandCorrectionCoefficientA']
                    b_ir = rads.attrs['BandCorrectionCoefficientB']
                    lambda_c = rads.attrs['CentralWaveLength']
                    arr *= 1e6
                    arr = (h * c) / (k * lambda_c * \
                                     np.log(1 +
                                            (2 * h * c ** 2) /
                                            ((lambda_c ** 5) * arr)))
                    arr *= a_ir
                    arr += b_ir
                    unit = "K"
                except KeyError:
                    logger.warning("Calibration failed.")

        elif calibrate != 2:
            raise ValueError("Calibrate parameter should be 1 or 2")
        arr[arr < 0] = 0
        res.append(arr)
        units.append(unit)

    return res, units


def read_dnb(h5f):

    scans = h5f.get_node("/All_Data/NumberOfScans").read()[0]
    res = []
    units = []

    rads_dset = h5f.get_node("/All_Data/VIIRS-DNB-SDR_All")
    arr = np.ma.masked_greater(rads_dset.Radiance.read()[:scans * 16, :], 1.0)
    unit = "W m-2 sr-1 μm-1"
    arr[arr < 0] = 0
    res.append(arr)
    units.append(unit)

    return res, units


def expand_array(data, scans, c_align, c_exp, scan_size=16, tpz_size=16, nties=200, track_offset=0.5, scan_offset=0.5):
    s_track, s_scan = np.mgrid[0:scans * scan_size, 0:nties*tpz_size]
    s_track = (s_track.reshape(scans, scan_size, nties, tpz_size) % scan_size + track_offset) / scan_size
    s_scan = (s_scan.reshape(scans, scan_size, nties, tpz_size) % tpz_size + scan_offset) / tpz_size

    a_scan = s_scan + s_scan * (1 - s_scan) * c_exp + s_track * (1 - s_track) * c_align
    a_track = s_track

    data_a = data[:scans * 2:2, np.newaxis, :-1, np.newaxis]
    data_b = data[:scans * 2:2, np.newaxis, 1:, np.newaxis]
    data_c = data[1:scans * 2:2, np.newaxis, 1:, np.newaxis]
    data_d = data[1:scans * 2:2, np.newaxis, :-1, np.newaxis]

    fdata = ((1 - a_track) * ((1 - a_scan) * data_a + a_scan * data_b) +
             a_track * ((1 - a_scan) * data_d + a_scan * data_c))
    return fdata.reshape(scans * scan_size, nties * tpz_size)


def lonlat2xyz(lon, lat):
    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)
    x = np.cos(lat) * np.cos(lon)
    y = np.cos(lat) * np.sin(lon)
    z = np.sin(lat)
    return x, y, z


def xyz2lonlat(x, y, z):
    lon = np.rad2deg(np.arctan2(y, x))
    lat = np.rad2deg(np.arctan2(z, np.sqrt(x ** 2 + y ** 2)))
    return lon, lat


def navigate_m(h5f, channel):

    if not channel.startswith("M"):
        raise ValueError("Unknow channel type for band %s", channel)

    scans = h5f["All_Data"]["NumberOfScans"][0]
    geostuff = h5f["All_Data"]["VIIRS-MOD-GEO_All"]
    all_c_align = geostuff["AlignmentCoefficient"].value[np.newaxis, np.newaxis,
                                                     :, np.newaxis]
    all_c_exp = geostuff["ExpansionCoefficient"].value[np.newaxis, np.newaxis,
                                                   :, np.newaxis]
    all_lon = geostuff["Longitude"].value
    all_lat = geostuff["Latitude"].value

    res = []

    # FIXME:  this supposes  there is  only one  tiepoint zone  in the
    # track direction
    scan_size = h5f["All_Data/VIIRS-%s-SDR_All" % \
                    channel].attrs["TiePointZoneSizeTrack"][0]
    track_offset = h5f["All_Data/VIIRS-%s-SDR_All" % \
                       channel].attrs["PixelOffsetTrack"]
    scan_offset = h5f["All_Data/VIIRS-%s-SDR_All" % \
                      channel].attrs["PixelOffsetScan"]

    try:
        group_locations = h5f["All_Data/VIIRS-MOD-GEO_All/"
                              "TiePointZoneGroupLocationScanCompact"].value
    except KeyError:
        group_locations = [0]
    param_start = 0
    for tpz_size, nb_tpz, start in \
        zip(h5f["All_Data/VIIRS-%s-SDR_All" % \
                channel].attrs["TiePointZoneSizeScan"],
            h5f["All_Data/VIIRS-MOD-GEO_All/NumberOfTiePointZonesScan"].value,
            group_locations):
        lon = all_lon[:, start:start + nb_tpz + 1]
        lat = all_lat[:, start:start + nb_tpz + 1]
        c_align = all_c_align[:, :, param_start:param_start + nb_tpz, :]
        c_exp = all_c_exp[:, :, param_start:param_start + nb_tpz, :]
        param_start += nb_tpz
        nties = nb_tpz
        if (np.max(lon) - np.min(lon) > 90) or (np.max(abs(lat)) > 60):
            x, y, z = lonlat2xyz(lon, lat)
            x, y, z = (expand_array(x, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset, scan_offset),
                       expand_array(y, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset, scan_offset),
                       expand_array(z, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset, scan_offset))
            res.append(xyz2lonlat(x, y, z))
        else:
            res.append((expand_array(lon, scans, c_align, c_exp, scan_size,
                                     tpz_size, nties, track_offset,
                                     scan_offset),
                       expand_array(lat, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset,
                                    scan_offset)))
    lons, lats = zip(*res)
    return np.hstack(lons), np.hstack(lats)


def navigate_dnb(h5f):

    scans = h5f.get_node("/All_Data/NumberOfScans").read()[0]
    geo_dset = h5f.get_node("/All_Data/VIIRS-DNB-GEO_All")
    all_c_align = geo_dset.AlignmentCoefficient.read()[np.newaxis, np.newaxis,
                                                       :, np.newaxis]
    all_c_exp = geo_dset.ExpansionCoefficient.read()[np.newaxis, np.newaxis,
                                                     :, np.newaxis]
    all_lon = geo_dset.Longitude.read()
    all_lat = geo_dset.Latitude.read()

    res = []

    # FIXME: this supposes there is only one tiepoint zone in the
    # track direction
    scan_size = h5f.get_node_attr("/All_Data/VIIRS-DNB-SDR_All",
                                  "TiePointZoneSizeTrack")[0]
    track_offset = h5f.get_node_attr("/All_Data/VIIRS-DNB-SDR_All",
                                     "PixelOffsetTrack")[0]
    scan_offset = h5f.get_node_attr("/All_Data/VIIRS-DNB-SDR_All",
                                    "PixelOffsetScan")[0]

    try:
        group_locations = geo_dset.TiePointZoneGroupLocationScanCompact.read()
    except KeyError:
        group_locations = [0]
    param_start = 0
    for tpz_size, nb_tpz, start in \
        zip(h5f.get_node_attr("/All_Data/VIIRS-DNB-SDR_All",
                              "TiePointZoneSizeScan"),
            geo_dset.NumberOfTiePointZonesScan.read(),
            group_locations):
        lon = all_lon[:, start:start + nb_tpz + 1]
        lat = all_lat[:, start:start + nb_tpz + 1]
        c_align = all_c_align[:, :, param_start:param_start + nb_tpz, :]
        c_exp = all_c_exp[:, :, param_start:param_start + nb_tpz, :]
        param_start += nb_tpz
        nties = nb_tpz
        if (np.max(lon) - np.min(lon) > 90) or (np.max(abs(lat)) > 60):
            x, y, z = lonlat2xyz(lon, lat)
            x, y, z = (expand_array(x, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset, scan_offset),
                       expand_array(y, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset, scan_offset),
                       expand_array(z, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset, scan_offset))
            res.append(xyz2lonlat(x, y, z))
        else:
            res.append((expand_array(lon, scans, c_align, c_exp, scan_size,
                                     tpz_size, nties, track_offset,
                                     scan_offset),
                       expand_array(lat, scans, c_align, c_exp, scan_size,
                                    tpz_size, nties, track_offset,
                                    scan_offset)))
    lons, lats = zip(*res)
    return np.hstack(lons), np.hstack(lats)


if __name__ == '__main__':
    #filename = "/local_disk/data/satellite/polar/compact_viirs/SVMC_npp_d20140114_t1245125_e1246367_b11480_c20140114125427496143_eum_ops.h5"
    filename = "/local_disk/data/satellite/polar/compact_viirs/mymy.h5"
    h5f = h5py.File(filename, 'r')
    # ch1, ch2, ch3, ch4 = read(h5f, ["M5", "M4", "M2", "M12"])
    # img = GeoImage((ch1, ch2, ch3),
    #                None,
    #                None,
    #                fill_value=None,
    #                mode="RGB")

    # img.enhance(stretch="linear")
    # img.enhance(gamma=2.0)
    # img.show()

    lons, lats = navigate_m(h5f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017.

# Author(s):

#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""This modules describes the viirs instrument.
It provides VIIRS specific methods for RGB-compositing.
"""
import logging

import numpy as np

import mpop.imageo.geo_image as geo_image
from mpop.instruments.visir import VisirCompositer

LOG = logging.getLogger(__name__)

try:
    from pyorbital.astronomy import sun_zenith_angle as sza
except ImportError:
    LOG.warning("Sun zenith angle correction not possible! " +
                "Check the availability of the pyorbital module in your environment")
    sza = None


# VIIRS
# Since there is overlap between I-bands and M-bands we need to
# specifically re-define some of the RGB composites already defined
# in the standard visir.py module. So, the same names, like "overview"
# can be invoked and based on M-bands only.
# In addition we define new composite names for the I-bands,
# like e.g. hr_overview, hr_night_fog, etc
#


class ViirsCompositer(VisirCompositer):

    """This class sets up the VIIRS instrument channel list.
    """

    instrument_name = "viirs"

    def overview(self, stretch='linear', gamma=1.6):
        """Make an Overview RGB image composite from VIIRS
        channels.
        """
        self.check_channels('M05', 'M07', 'M15')

        ch1 = self['M05'].check_range()
        ch2 = self['M07'].check_range()
        ch3 = -self['M15'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")
        if stretch:
            img.enhance(stretch=stretch)
        if gamma:
            img.enhance(gamma=gamma)

        return img

    overview.prerequisites = set(['M05', 'M07', 'M15'])

    def overview_sun(self, stretch='linear', gamma=1.6, fill_value=(0, 0, 0)):
        """Make an overview RGB image composite normalising with cosine to the
        sun zenith angle.
        """
        self.check_channels('M05', 'M07', 'M15')

        lonlats = self['M15'].area.get_lonlats()

        red = self['M05'].sunzen_corr(self.time_slot, lonlats, limit=88.,
                                      sunmask=95).data
        green = self['M07'].sunzen_corr(self.time_slot, lonlats, limit=88.,
                                        sunmask=95).data
        blue = -self['M15'].data

        img = geo_image.GeoImage((red, green, blue),
                                 self.area,
                                 self.time_slot,
                                 fill_value=fill_value,
                                 mode="RGB")

        if stretch:
            img.enhance(stretch=stretch)
        if gamma:
            img.enhance(gamma=gamma)

        return img

    overview_sun.prerequisites = set(['M05', 'M07', 'M15'])

    def hr_overview(self):
        """Make a high resolution Overview RGB image composite 
        from the VIIRS I-bands only - 375 meter resolution.
        """
        self.check_channels('I01', 'I02', 'I05')

        ch1 = self['I01'].check_range()
        ch2 = self['I02'].check_range()
        ch3 = -self['I05'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")

        img.enhance(stretch="crude")
        img.enhance(gamma=1.6)

        return img

    hr_overview.prerequisites = set(['I01', 'I02', 'I05'])

    def truecolor(self, stretch='linear', gamma=2.0):
        """Make a True Color RGB image composite from
        M-bands only.
        """
        #elf.check_channels('M02', 'M04', 'M05')
        self.check_channels('M03', 'M04', 'M05')

        ch1 = self['M05'].check_range()
        ch2 = self['M04'].check_range()
        #h3 = self['M02'].check_range()
        ch3 = self['M03'].check_range()

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=None,
                                 mode="RGB")

        if stretch:
            img.enhance(stretch=stretch)
        if gamma:
            img.enhance(gamma=gamma)

        return img

    truecolor.prerequisites = set(['M03', 'M04', 'M05'])
#   truecolor.prerequisites = set(['M02', 'M04', 'M05'])

    def natural(self):
        """Make a Natural Colors RGB image composite from
        M-bands only.
        """
        self.check_channels('M05', 'M06', 'M07', 'M10')

        ch1 = self['M10'].check_range()
        ch2 = self['M07'].check_range()
        ch3 = self['M05'].check_range()

        ch2b = self['M06'].check_range()
        ch2 = np.ma.where(ch2.mask, ch2b, ch2)

        common_mask = np.logical_or(ch1.mask, ch2.mask)
        common_mask = np.logical_or(common_mask, ch3.mask)
        ch1.mask = common_mask
        ch2.mask = common_mask
        ch3.mask = common_mask

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB",
                                 crange=((0, 90),
                                         (0, 90),
                                         (0, 90)))

        img.enhance(gamma=1.8)

        return img

    natural.prerequisites = set(['M05', 'M06', 'M07', 'M10'])

    def hr_natural(self):
        """Make a high resolution Day Natural Colors RGB image 
        composite from I-bands only - 375 meter resolution.
        """
        self.check_channels('I01', 'I02', 'I03')

        ch1 = self['I03'].check_range()
        ch2 = self['I02'].check_range()
        ch3 = self['I01'].check_range()

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB",
                                 crange=((0, 90),
                                         (0, 90),
                                         (0, 90)))

        img.enhance(gamma=1.8)

        return img

    hr_natural.prerequisites = set(['I01', 'I02', 'I03'])

    def vis06(self):
        """Make a black and white image of the VIS 0.635um channel.
        """
        return self.channel_image("M05")

    vis06.prerequisites = set(['M05'])

    def hr_vis06(self):
        """Make a high res black and white image of the 
        'visible' (VIS) I-band at 0.640um.
        """
        return self.channel_image('I01')

    hr_vis06.prerequisites = set(['I01'])

    def green_snow(self):
        """Make a Green Snow RGB image composite.
        """
        self.check_channels('M05', 'M10', 'M15')

        ch1 = self['M10'].check_range()
        ch2 = self['M05'].check_range()
        ch3 = -self['M15'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")

        img.enhance(stretch="crude")
        img.enhance(gamma=1.6)

        return img

    green_snow.prerequisites = set(['M05', 'M10', 'M15'])

    def hr_green_snow(self):
        """Make a Green Snow RGB image composite.
        """
        self.check_channels('I01', 'I03', 'I05')

        ch1 = self['I03'].check_range()
        ch2 = self['I01'].check_range()
        ch3 = -self['I05'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")

        img.enhance(stretch="crude")
        img.enhance(gamma=1.6)

        return img

    hr_green_snow.prerequisites = set(['I01', 'I03', 'I05'])

    def red_snow(self):
        """Make a Red Snow RGB image composite.
        """
        self.check_channels('M05', 'M10', 'M15')

        ch1 = self['M05'].check_range()
        ch2 = self['M10'].check_range()
        ch3 = -self['M15'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")

        img.enhance(stretch="crude")

        return img

    red_snow.prerequisites = set(['M05', 'M10', 'M15'])

    def hr_red_snow(self):
        """Make a high resolution Red Snow RGB image composite
        from the I-bands only.
        """
        self.check_channels('I01', 'I03', 'I05')

        ch1 = self['I01'].check_range()
        ch2 = self['I03'].check_range()
        ch3 = -self['I05'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")

        img.enhance(stretch="crude")

        return img

    hr_red_snow.prerequisites = set(['I01', 'I03', 'I05'])

    def dnb_overview(self, stretch='linear'):
        """Make a nighttime overview RGB image composite from VIIRS
        DNB and M bands.
        """
        self.check_channels('DNB', 'M15')

        lonlats = self['M15'].area.get_lonlats()

        if sza:
            sunz = sza(self.time_slot, lonlats[0], lonlats[1])
            sunz = np.ma.masked_outside(sunz, 103, 180)
            sunzmask = sunz.mask

            red = np.ma.masked_where(sunzmask, self['DNB'].data)
            green = np.ma.masked_where(sunzmask, self['DNB'].data)
            blue = np.ma.masked_where(sunzmask, -self['M15'].data)
        else:
            LOG.warning("No masking of solar contaminated pixels performed!")
            red = self['DNB'].data
            green = self['DNB'].data
            blue = -self['M15'].data

        img = geo_image.GeoImage((red, green, blue),
                                 self.area,
                                 self.time_slot,
                                 fill_value=None,
                                 mode="RGB")

        img.enhance(stretch=stretch)

        return img

    dnb_overview.prerequisites = set(['DNB', 'M15'])

    # def dnb_overview(self):
    #     """Make an Overview RGB image composite from VIIRS
    #     channels.
    #     """
    #     self.check_channels('DNB', 'M15')

    #     ch1 = self['DNB'].data
    #     ch2 = self['DNB'].data
    #     ch3 = -self['M15'].data

    #     img = geo_image.GeoImage((ch1, ch2, ch3),
    #                              self.area,
    #                              self.time_slot,
    #                              fill_value=None,
    #                              mode="RGB")

    #     img.enhance(stretch="linear")

    #     return img

    # dnb_overview.prerequisites = set(['DNB', 'M15'])

    def night_color(self):
        """Make a Night Overview RGB image composite.
        Same as cloudtop ... just different.
        """
        return self.cloudtop(stretch="histogram")

    night_color.prerequisites = set(['M12', 'M15', 'M16'])

    def night_fog(self):
        """Make a Night Fog RGB image composite.
        """
        self.check_channels('M12', 'M15', 'M16')

        ch1 = self['M16'].data - self['M15'].data
        ch2 = self['M15'].data - self['M12'].data
        ch3 = self['M15'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB",
                                 crange=((-4, 2),
                                         (0, 6),
                                         (243, 293)))

        img.enhance(gamma=(1.0, 2.0, 1.0))

        return img

    night_fog.prerequisites = set(['M12', 'M15', 'M16'])

    def dust(self):
        """Make a Dust RGB image composite.
        """
        self.check_channels('M14', 'M15', 'M16')

        ch1 = self['M16'].data - self['M15'].data
        ch2 = self['M15'].data - self['M14'].data
        ch3 = self['M15'].data
        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB",
                                 crange=((-4, 2),
                                         (0, 15),
                                         (261, 289)))

        img.enhance(gamma=(1.0, 2.5, 1.0))

        return img

    dust.prerequisites = set(['M14', 'M15', 'M16'])

    def ash(self):
        """Make a Ash RGB image composite.
        """
        self.check_channels('M14', 'M15', 'M16')

        ch1 = self['M16'].data - self['M15'].data
        ch2 = self['M15'].data - self['M14'].data
        ch3 = self['M15'].data
        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB",
                                 crange=((-4, 2),
                                         (-4, 5),
                                         (243, 303)))

        return img

    ash.prerequisites = set(['M14', 'M15', 'M16'])

    def fog(self):
        """Make a Fog RGB image composite.
        """
        self.check_channels('M14', 'M15', 'M16')

        ch1 = self['M16'].data - self['M15'].data
        ch2 = self['M15'].data - self['M14'].data
        ch3 = self['M15'].data
        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB",
                                 crange=((-4, 2),
                                         (0, 6),
                                         (243, 283)))

        img.enhance(gamma=(1.0, 2.0, 1.0))

        return img

    fog.prerequisites = set(['M14', 'M15', 'M16'])

    def cloudtop(self, stretch=None):
        """Make a Cloudtop RGB image composite.
        """
        self.check_channels('M12', 'M15', 'M16')

        ch1 = -self['M12'].data
        ch2 = -self['M15'].data
        ch3 = -self['M16'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")

        if stretch:
            img.enhance(stretch=stretch)
        else:
            img.enhance(stretch=(0.005, 0.005))

        return img

    cloudtop.prerequisites = set(['M12', 'M15', 'M16'])

    def dnb(self, stretch="histogram"):
        """Make a black and white image of the Day-Night band."""
        self.check_channels('DNB')

        img = geo_image.GeoImage(self['DNB'].data,
                                 self.area,
                                 self.time_slot,
                                 fill_value=0,
                                 mode="L")
        if stretch:
            img.enhance(stretch=stretch)
        return img

    dnb.prerequisites = set(['DNB'])

    def dnb_rgb(self, stretch="linear"):
        """Make a RGB Day-Night band using M15 as blue."""
        self.check_channels('DNB', 'M15')
        ch1 = self['DNB'].data
        ch2 = self['DNB'].data
        ch3 = -self['M15'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")
        if stretch:
            img.enhance(stretch=stretch)
        return img

    dnb_rgb.prerequisites = set(['DNB', 'M15'])

    def ir108(self):
        """Make a black and white image of the IR 10.8um channel.
        """
        self.check_channels("M15")

        img = geo_image.GeoImage(self["M15"].data,
                                 self.area,
                                 self.time_slot,
                                 fill_value=0,
                                 mode="L",
                                 crange=(-70 + 273.15, 57.5 + 273.15))
        img.enhance(inverse=True)
        return img

    ir108.prerequisites = set(["M15"])

    def hr_ir108(self):
        """Make a black and white image of the IR 10.8um channel (320m).
        """
        self.check_channels("I05")

        img = geo_image.GeoImage(self["I05"].data,
                                 self.area,
                                 self.time_slot,
                                 fill_value=0,
                                 mode="L",
                                 crange=(-70 + 273.15, 57.5 + 273.15))
        img.enhance(inverse=True)
        return img

    hr_ir108.prerequisites = set(["I05"])

    def chlorophyll(self, stretch=None):
        """ From http://oceancolor.gsfc.nasa.gov/REPROCESSING/R2009/ocv6/

        * Rrs1 = blue wavelength Rrs (e.g., 443, 490, or 510-nm)
        * Rrs2 = green wavelength Rrs (e.g., 547, 555, or 565-nm)
        * X = log10(Rrs1 / Rrs2)
        * chlor_a = 10^(a0 + a1*X + a2*X^2 + a3*X^3 + a4*X^4)

        sensor  default *      blue     green   a0       a1      a2       a3       a4
        OC3V    VIIRS   Y      443>486  550     0.2228  -2.4683  1.5867  -0.4275  -0.7768

        blue: M02(445)>M03(488)
        green: M04(555)

        * X = log10(max(M2, M3)/M4)
        """
        self.check_channels("M02", "M03", "M04")

        a0, a1, a2, a3, a4 = (0.2228, -2.4683, 1.5867, -0.4275, -0.7768)

        # X = np.maximum(self["M02"].data, self["M03"].data)/self["M04"].data
        X = self["M02"].data / self["M04"].data
        X = np.log10(X)
        chlor_a = 10 ** (a0 + a1 * X + a2 * (X ** 2) +
                         a3 * (X ** 3) + a4 * (X ** 4))
        print 'chlor_a:', chlor_a.min(), chlor_a.mean(), chlor_a.max()

        img = geo_image.GeoImage(chlor_a,
                                 self.area,
                                 self.time_slot,
                                 fill_value=0,
                                 mode="L")

        if stretch:
            img.enhance(stretch=stretch)
        return img

    chlorophyll.prerequisites = set(["M02", "M03", "M04"])

    def hr_cloudtop(self):
        """Make a Night Fog RGB image composite.
        """
        self.check_channels('I04', 'I05')

        ch1 = -self['I04'].data
        ch2 = self['I05'].data
        ch3 = self['I05'].data

        img = geo_image.GeoImage((ch1, ch2, ch3),
                                 self.area,
                                 self.time_slot,
                                 fill_value=(0, 0, 0),
                                 mode="RGB")

        img.enhance(stretch=(0.005, 0.005))

        return img

    hr_cloudtop.prerequisites = set(['I04', 'I05'])

    def snow_age(self):
        """Make a Snow age RGB image composite.
        """
        self.check_channels('M07', 'M08', 'M09', 'M10', 'M11')

        coeff = 255. / 160.

        lonlats = self['M11'].area.get_lonlats()

        m07 = self['M07'].sunzen_corr(
            self.time_slot, lonlats, limit=88., sunmask=95).data * coeff
        m08 = self['M08'].sunzen_corr(
            self.time_slot, lonlats, limit=88., sunmask=95).data * coeff
        m09 = self['M09'].sunzen_corr(
            self.time_slot, lonlats, limit=88., sunmask=95).data * coeff
        m10 = self['M10'].sunzen_corr(
            self.time_slot, lonlats, limit=88., sunmask=95).data * coeff
        m11 = self['M11'].sunzen_corr(
            self.time_slot, lonlats, limit=88., sunmask=95).data * coeff

        refcu = m11 - m10
        refcu[refcu < 0] = 0

        ch1 = m07 - refcu / 2. - m09 / 4.
        ch2 = m08 + refcu / 4. + m09 / 4.
        ch3 = m11 + m09

        # Bernard Bellec snow Look-Up Tables V 1.0 (c) Meteo-France
        # These Look-up Tables allow you to create the RGB snow product
        # for SUOMI-NPP VIIRS Imager according to the algorithm
        # presented at the second CSPP/IMAPP users' meeting at Eumetsat
        # in Darmstadt on 14-16 April 2015
        # The algorithm and the product are described in this
        # presentation :
        # http://www.ssec.wisc.edu/meetings/cspp/2015/Agenda%20PDF/Wednesday/Roquet_snow_product_cspp2015.pdf
        #
        # For further information you may contact
        # Bernard Bellec at Bernard.Bellec@meteo.fr
        # or
        # Pascale Roquet at Pascale.Roquet@meteo.fr

        luts = np.array([[0, 0, 0], [1, 2, 2], [3, 8, 5], [4, 12, 8], [6, 15, 10], [8, 18, 13], [9, 21, 16],
                         [11, 24, 19], [13, 26, 21], [14, 28, 24], [
                             16, 30, 27], [18, 32, 30], [19, 34, 32],
                         [21, 36, 35], [22, 38, 38], [24, 40, 40], [
                             26, 42, 43], [27, 43, 46], [29, 45, 49],
                         [31, 47, 51], [32, 49, 54], [34, 50, 57], [
                             36, 52, 60], [37, 54, 62], [39, 55, 65],
                         [40, 57, 68], [42, 59, 70], [44, 60, 73], [
                             45, 62, 76], [47, 64, 79], [49, 66, 81],
                         [50, 67, 84], [52, 69, 87], [53, 71, 90], [
                             55, 73, 92], [56, 75, 95], [58, 77, 98],
                         [59, 79, 100], [61, 81, 103], [62, 83, 106], [
                             64, 85, 109], [65, 86, 111], [67, 88, 114],
                         [68, 90, 117], [70, 92, 119], [71, 94, 121], [
                             73, 96, 124], [74, 98, 126], [76, 100, 129],
                         [77, 102, 131], [79, 104, 134], [80, 106, 136], [
                             82, 107, 139], [83, 109, 141], [85, 111, 144],
                         [86, 113, 146], [88, 115, 149], [89, 117, 151], [
                             91, 118, 154], [92, 120, 156], [94, 122, 159],
                         [95, 124, 161], [97, 126, 162], [98, 128, 164], [
                             100, 129, 166], [101, 131, 168],
                         [103, 133, 170], [104, 135, 172], [106, 137, 173], [
                             107, 138, 175], [109, 140, 177],
                         [110, 142, 179], [112, 144, 181], [113, 145, 183], [
                             114, 147, 184], [116, 149, 186],
                         [117, 151, 188], [118, 152, 190], [120, 154, 192], [
                             121, 156, 193], [123, 158, 194],
                         [124, 159, 196], [125, 161, 197], [127, 163, 199], [
                             128, 165, 200], [130, 166, 202],
                         [131, 168, 203], [132, 170, 205], [134, 172, 206], [
                             135, 173, 206], [136, 175, 207],
                         [138, 177, 208], [139, 178, 209], [141, 180, 210], [
                             142, 182, 211], [143, 184, 212],
                         [145, 185, 213], [146, 187, 214], [148, 189, 215], [
                             149, 191, 216], [150, 192, 217],
                         [152, 194, 218], [153, 196, 219], [154, 198, 220], [
                             156, 200, 220], [157, 201, 221],
                         [159, 203, 221], [160, 205, 222], [161, 207, 223], [
                             162, 209, 223], [163, 210, 224],
                         [164, 212, 225], [166, 213, 225], [167, 214, 226], [
                             168, 216, 227], [169, 217, 227],
                         [171, 218, 228], [173, 220, 228], [174, 221, 228], [
                             175, 222, 229], [176, 224, 229],
                         [177, 225, 229], [178, 226, 230], [179, 227, 230], [
                             181, 228, 230], [182, 229, 231],
                         [183, 230, 231], [184, 231, 232], [185, 232, 232], [
                             186, 233, 232], [187, 234, 233],
                         [188, 235, 233], [190, 236, 233], [191, 237, 234], [
                             192, 237, 234], [193, 238, 234],
                         [194, 239, 235], [195, 240, 235], [196, 240, 236], [
                             196, 241, 236], [197, 242, 236],
                         [198, 243, 237], [199, 243, 237], [200, 244, 237], [
                             201, 245, 238], [202, 245, 238],
                         [203, 245, 238], [204, 246, 239], [205, 246, 239], [
                             206, 246, 239], [207, 247, 239],
                         [208, 247, 239], [209, 247, 239], [209, 248, 240], [
                             210, 248, 240], [210, 248, 240],
                         [211, 248, 240], [212, 248, 240], [212, 248, 241], [
                             213, 248, 241], [214, 248, 241],
                         [215, 248, 241], [216, 248, 241], [217, 248, 242], [
                             217, 248, 242], [218, 248, 242],
                         [219, 248, 242], [219, 248, 242], [220, 248, 243], [
                             221, 248, 243], [221, 249, 243],
                         [222, 249, 243], [223, 249, 243], [223, 249, 244], [
                             223, 249, 244], [224, 249, 244],
                         [224, 249, 244], [225, 249, 245], [225, 249, 245], [
                             226, 249, 245], [226, 249, 245],
                         [227, 249, 245], [227, 249, 246], [228, 249, 246], [
                             228, 250, 246], [229, 250, 246],
                         [229, 250, 246], [230, 250, 247], [230, 250, 247], [
                             231, 250, 247], [231, 250, 247],
                         [232, 250, 247], [233, 250, 248], [233, 250, 248], [
                             233, 250, 248], [234, 250, 248],
                         [234, 250, 248], [234, 250, 249], [235, 251, 249], [
                             235, 251, 249], [235, 251, 249],
                         [236, 251, 249], [236, 251, 250], [237, 251, 250], [
                             237, 251, 250], [237, 251, 250],
                         [238, 251, 250], [238, 251, 250], [238, 251, 250], [
                             239, 251, 250], [239, 251, 250],
                         [240, 251, 250], [240, 251, 250], [240, 252, 250], [
                             241, 252, 250], [241, 252, 251],
                         [241, 252, 251], [242, 252, 251], [242, 252, 251], [
                             242, 252, 251], [243, 252, 251],
                         [243, 252, 251], [244, 252, 251], [244, 252, 251], [
                             244, 252, 251], [245, 252, 252],
                         [245, 252, 252], [245, 253, 252], [246, 253, 252], [
                             246, 253, 252], [247, 253, 252],
                         [248, 253, 252], [248, 253, 252], [248, 253, 252], [
                             249, 253, 252], [249, 253, 253],
                         [249, 253, 253], [250, 253, 253], [250, 253, 253], [
                             250, 253, 253], [250, 253, 253],
                         [251, 254, 253], [251, 254, 253], [251, 254, 253], [
                             252, 254, 253], [252, 254, 254],
                         [252, 254, 254], [253, 254, 254], [253, 254, 254], [
                             253, 254, 254], [253, 254, 254],
                         [254, 254, 254], [254, 254, 254], [254, 254, 254], [254, 254, 254], [255, 255, 255]]) / 255.0
        np.ma.clip(ch1, 0, 255, ch1)
        np.ma.clip(ch2, 0, 255, ch2)
        np.ma.clip(ch3, 0, 255, ch3)
        ch1 = np.ma.array(
            luts[:, 0][ch1.astype(np.uint8)], copy=False, mask=ch1.mask)
        ch2 = np.ma.array(
            luts[:, 1][ch2.astype(np.uint8)], copy=False, mask=ch2.mask)
        ch3 = np.ma.array(
            luts[:, 2][ch3.astype(np.uint8)], copy=False, mask=ch3.mask)

        img = geo_image.GeoImage(
            (ch1, ch2, ch3), self.area, self.time_slot, mode="RGB")

        return img

    snow_age.prerequisites = set(['M07', 'M08', 'M09', 'M10', 'M11'])



#!/usr/bin/python

import xml.etree.ElementTree as ET
from pyresample import utils
import pickle
import urllib2

length=109800

#https://sentinel.esa.int/documents/247904/1955685/S2A_OPER_GIP_TILPAR_20150622T000000_21000101T000000_ZZ_0001
FNAME="S2A_OPER_GIP_TILPAR_20150622T000000_21000101T000000_ZZ_0001"

TILPAR_URL="https://sentinel.esa.int/documents/247904/1955685/"+FNAME

FNAME=FNAME+".kml"

tiles = urllib2.urlopen(TILPAR_URL)
with open(FNAME,'wb') as output:
  output.write(tiles.read())

tiles.close()

tree = ET.parse(FNAME)
root = tree.getroot()

s2tiles={}

for pm in root.iter('{http://www.opengis.net/kml/2.2}Placemark'):
	tilename=None
	epsg=None
	utm_ul_x=None
	utm_ul_y=None

	for name in pm.iter('{http://www.opengis.net/kml/2.2}name'):
		tilename=name.text
	for simple in pm.iter('{http://www.opengis.net/kml/2.2}SimpleData'):
		if (simple.attrib['name']=='epsg'):
			epsg=simple.text
		if(simple.attrib['name']=='utm_ul_x'):
			utm_ul_x=simple.text
		if(simple.attrib['name']=='utm_ul_y'):
			utm_ul_y=simple.text

	extent=(float(utm_ul_x),float(utm_ul_y)-length,float(utm_ul_x)+length,float(utm_ul_y))

	s2tiles[tilename]=[epsg,extent]

f=open('s2tiles.pickle','w')
pickle.dump(s2tiles,f)
f.close()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The tests package.
"""

from mpop.tests import (test_satin_helpers,
                        #test_pp_core, # crash
                        test_channel,
                        test_image,
                        test_geo_image,
                        #test_mipp,
                        test_projector,
                        #test_satellites,
                        test_scene,
                        #test_seviri,
                        #test_viirs_sdr,
                        #test_visir,
                        )

import unittest

def suite():
    """The global test suite.
    """

    mysuite = unittest.TestSuite()
    mysuite.addTests(test_satin_helpers.suite())
    #mysuite.addTests(test_pp_core.suite())
    mysuite.addTests(test_channel.suite())
    mysuite.addTests(test_image.suite())
    mysuite.addTests(test_geo_image.suite())
    #mysuite.addTests(test_mipp.suite())
    mysuite.addTests(test_projector.suite())
    #mysuite.addTests(test_satellites.suite())
    mysuite.addTests(test_scene.suite())
    #mysuite.addTests(test_seviri.suite())
    #mysuite.addTests(test_viirs_sdr.suite())
    #mysuite.addTests(test_visir.suite())

    return mysuite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2011, 2013.

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Module to compute satellite positions from TLE.
"""
import datetime
import sys
import urllib2

import numpy as np


CK2 = 5.413080e-4
CK4 = 0.62098875e-6
E6A = 1.0e-6
QOMS2T = 1.88027916e-9
S = 1.01222928
XJ3 = -0.253881e-5
XKE = 0.743669161e-1
XKMPER = 6378.137
XMNPDA = 1440.0
AE = 1.0
# earth flattening
F = 1/298.257223563


if sys.version_info < (2, 5):
    import time
    def strptime(string, fmt=None):
        """This function is available in the datetime module only
        from Python >= 2.5.
        """
        
        return datetime.datetime(*time.strptime(string, fmt)[:6])
else:
    strptime = datetime.datetime.strptime
class Tle(object):
    """The TLE object holds information and methods for orbit position
    estimation.
    """


    def __init__(self, tle=None, satellite=None):
        self.tle = tle
        if satellite:
            tles_dict = {}

            import glob
            filelist = glob.glob("/data/24/saf/polar_in/tle/tle-*.txt")
            if len(filelist) > 0:
                filelist.sort()
                tlef = open(filelist[-1])
                tles = [item.strip() for item in tlef]
                tlef.close()
                for i in xrange(0, len(tles) - 2, 3):
                    tles_dict[tles[i]] = tles[i+1]+"\n"+tles[i+2]
            else:
                for fname in ["resource.txt", "weather.txt"]:
                    url = "http://celestrak.com/NORAD/elements/" + fname
                    tles = urllib2.urlopen(url).readlines()
                    tles = [item.strip() for item in tles]

                    for i in xrange(0, len(tles) - 2, 3):
                        tles_dict[tles[i]] = tles[i+1]+"\n"+tles[i+2]

            self._read_tle(tles_dict[satellite.upper()])
            self._preprocess()


    def _read_tle(self, lines):
        """Read the raw tle.
        """

        def _read_tle_decimal(rep):
            """Read tle decimal point numbers.
            """
            num = int(rep[:-2]) * 1.0e-5
            exp = int(rep[-2:])
            return num * 10 ** exp

        tlist = lines.split()
        self.tle = {}
        self.tle["satnumber"] = tlist[1][:5]
        self.tle["classification"] = tlist[1][5:]
        self.tle["id_launch_year"] = tlist[2][:2]
        self.tle["id_launch_number"] = tlist[2][2:5]
        self.tle["id_launch_piece"] = tlist[2][5:]
        self.tle["epoch_year"] = int(tlist[3][:2])
        self.tle["epoch_day"] = float(tlist[3][2:])
        self.tle["epoch"] = (strptime(tlist[3][:2], "%y") +
                        datetime.timedelta(days=float(tlist[3][2:]) - 1))
        self.tle["mean_motion_derivative"] = float(tlist[4])
        self.tle["mean_motion_sec_derivative"] = _read_tle_decimal(tlist[5])
        self.tle["bstar"] = _read_tle_decimal(tlist[6])
        self.tle["ephemeris_type"] = int(tlist[7])
        self.tle["element_number"] = int(tlist[8][:-1])

        self.tle["inclination"] = float(tlist[11])
        self.tle["right_ascension"] = float(tlist[12])
        self.tle["excentricity"] = int(tlist[13]) * 10 ** -7
        self.tle["arg_perigee"] = float(tlist[14])
        self.tle["mean_anomaly"] = float(tlist[15])
        self.tle["mean_motion"] = float(tlist[16][:11])
        self.tle["orbit"] = int(tlist[16][11:-1])


    def _preprocess(self):
        """Derivate some values from raw tle.
        """
        self.tle["inclination"] = np.deg2rad(self.tle["inclination"])
        self.tle["right_ascension"] = np.deg2rad(self.tle["right_ascension"])
        self.tle["arg_perigee"] = np.deg2rad(self.tle["arg_perigee"])
        self.tle["mean_anomaly"] = np.deg2rad(self.tle["mean_anomaly"])

        self.tle["mean_motion"] *= (np.pi * 2 / XMNPDA)
        self.tle["mean_motion_derivative"] *= np.pi * 2 / XMNPDA ** 2
        self.tle["mean_motion_sec_derivative"] *= np.pi * 2 / XMNPDA ** 3
        self.tle["bstar"] *= AE

        n_0 = self.tle["mean_motion"]
        k_e = XKE
        k_2 = CK2
        i_0 = self.tle["inclination"]
        e_0 = self.tle["excentricity"]

        a_1 = (k_e / n_0) ** (2.0/3)
        delta_1 = ((3/2.0) * (k_2 / a_1**2) * ((3 * np.cos(i_0)**2 - 1) /
                                              (1 - e_0**2)**(2.0/3)))

        a_0 = a_1 * (1 - delta_1/3 - delta_1**2 - (134.0/81) * delta_1**3)

        delta_0 = ((3/2.0) * (k_2 / a_0**2) * ((3 * np.cos(i_0)**2 - 1) /
                                              (1 - e_0**2)**(2.0/3)))

        # original mean motion
        n_0pp = n_0 / (1 + delta_0)
        self.tle["original_mean_motion"] = n_0pp

        # semi major axis
        a_0pp = a_0 / (1 - delta_0)
        self.tle["semi_major_axis"] = a_0pp

        self.tle["period"] = np.pi * 2 / n_0pp

        self.tle["perigee"] = (a_0pp * (1 - e_0) / AE - AE) * XKMPER

        now = self.tle["epoch"]

        self.tle["right_ascension_lon"] = (self.tle["right_ascension"]
                                           - gmst(now))

        if self.tle["right_ascension_lon"] > np.pi:
            self.tle["right_ascension_lon"] -= 2 * np.pi

# pylint: disable-msg=C0103

    def get_position(self, current_time):
        """Get cartesian position and velocity.
        """
        # for near earth orbits, period must be < 255 minutes

        perigee = self.tle["perigee"]
        a_0pp = self.tle["semi_major_axis"]
        e_0 = self.tle["excentricity"]
        i_0 = self.tle["inclination"]
        n_0pp = self.tle["original_mean_motion"]
        k_2 = CK2
        k_4 = CK4
        k_e = XKE
        bstar = self.tle["bstar"]
        w_0 = self.tle["arg_perigee"]
        M_0 = self.tle["mean_anomaly"]
        W_0 = self.tle["right_ascension"]
        t_0 = self.tle["epoch"]
        A30 = -XJ3 * AE**3

        if perigee < 98:
            s = 20/XKMPER + AE
            qoms2t = (QOMS2T ** 0.25 + S - s) ** 4
        elif perigee < 156:
            s = a_0pp * (1 - e_0) - S + AE 
            qoms2t = (QOMS2T ** 0.25 + S - s) ** 4
        else:
            qoms2t = QOMS2T
            s = S

        theta = np.cos(i_0)
        xi = 1 / (a_0pp - s)
        beta_0 = np.sqrt(1 - e_0 ** 2)
        eta = a_0pp * e_0 * xi

        C_2 = (qoms2t * xi**4 * n_0pp * (1 - eta**2)**(-3.5) *
               (a_0pp * (1 + 1.5 * eta**2 + 4 * e_0 * eta + e_0 * eta**3) +
                1.5 * (k_2 * xi) / (1 - eta**2) * (-0.5 + 1.5 * theta**2)*
                (8 + 24 * eta**2 + 3 * eta**4)))

        C_1 = bstar * C_2

        C_3 = (qoms2t * xi ** 5 * A30 * n_0pp * AE * np.sin(i_0) / (k_2 * e_0))

        coef = 2 * qoms2t * xi**4 * a_0pp * beta_0**2*(1-eta**2)**(-7/2.0)

        C_4 = (coef * n_0pp *
               ((2 * eta * (1 + e_0 * eta) + e_0/2.0 + (eta**3)/2.0) -
                2 * k_2 * xi / (a_0pp * (1 - eta**2)) *
                (3*(1-3*theta**2) *
                 (1 + (3*eta**2)/2.0 - 2*e_0*eta - e_0*eta**3/2.0) +
                 3/4.0*(1-theta**2)*
                 (2*eta**2 - e_0*eta - e_0*eta**3)*np.cos(2*w_0))))

        C_5 = coef * (1 + 11/4.0 * eta * (eta + e_0) + e_0 * eta**3)
        D_2 = 4 * a_0pp * xi * C_1**2
        D_3 = 4/3.0 * a_0pp * xi**2 * (17*a_0pp + s) * C_1**3
        D_4 = 2/3.0 * a_0pp * xi**3 * (221*a_0pp + 31*s) * C_1**4

        # Secular effects of atmospheric drag and gravitation
        dt = _days(current_time - t_0) * XMNPDA

        M_df = (M_0 + (1 +
                       3*k_2*(-1 + 3*theta**2)/(2*a_0pp**2 * beta_0**3) +
                       3*k_2**2*(13 - 78*theta**2 + 137*theta**4)/
                       (16*a_0pp**4*beta_0**7))*
                n_0pp*dt)
        w_df = (w_0 + (-3*k_2*(1 - 5*theta**2)/(2*a_0pp**2*beta_0**4) +
                       3 * k_2**2 * (7 - 114*theta**2 + 395*theta**4)/
                       (16*a_0pp*beta_0**8) +
                       5*k_4*(3-36*theta**2+49*theta**4)/
                       (4*a_0pp**4*beta_0**8))*
                n_0pp*dt)
        W_df = (W_0 + (-3*k_2*theta/(a_0pp**2*beta_0**4) +
                       3*k_2**2*(4*theta- 19*theta**3)/(2*a_0pp**4*beta_0**8) +
                       5*k_4*theta*(3-7*theta**2)/(2*a_0pp**4*beta_0**8))*
                n_0pp*dt)
        deltaw = bstar * C_3 * np.cos(w_0)*dt
        deltaM = (-2/3.0 * qoms2t * bstar * xi**4 * AE / (e_0*eta) *
                  ((1 + eta * np.cos(M_df))**3 - (1 + eta * np.cos(M_0))**3))
        M_p = M_df + deltaw + deltaM
        w = w_df - deltaw - deltaM
        W = (W_df - 21/2.0 * (n_0pp * k_2 * theta)/(a_0pp**2 * beta_0**2) *
             C_1 * dt**2)

        e = (e_0 -
             bstar * C_4 * dt -
             bstar * C_5 * (np.sin(M_p) - np.sin(M_0)))

        a = a_0pp * (1 - C_1 * dt - D_2 * dt**2 - D_3 * dt**3 - D_4 * dt**4)**2
        L = M_p + w + W + n_0pp * (3/2.0 * C_1 * dt**2 +
                                   (D_2 + 2 * C_1 ** 2) * dt**3 +
                                   1/4.0 *
                                   (3*D_3 + 12*C_1*D_2 + 10*C_1**3)*dt**4 +
                                   1.0/5 * (3*D_4 + 12*C_1*D_3 + 6*D_2**2 +
                                            30*C_1**2*D_2 + 15*C_1**4)*dt**5)
        beta = np.sqrt(1 - e**2)
        n = k_e / (a ** (3/2.0))

        # Long-period periodic terms
        a_xN = e * np.cos(w)
        a_yNL = A30 * np.sin(i_0) / (4.0 * k_2 * a * beta**2)
        L_L = a_yNL/2 * a_xN * ((3 + 5 * theta) / (1 + theta))
        L_T = L + L_L
        a_yN = e * np.sin(w) + a_yNL

        U = (L_T - W) % (np.pi * 2)

        Epw = U
        for i in range(10):
            DeltaEpw = ((U - a_yN * np.cos(Epw) + a_xN  * np.sin(Epw) - Epw) /
                        (-a_yN * np.sin(Epw) - a_xN * np.cos(Epw) + 1))
            Epw = Epw + DeltaEpw
            if DeltaEpw < 10e-12:
                break

        # preliminary quantities for short-period periodics

        ecosE = a_xN * np.cos(Epw) + a_yN * np.sin(Epw)
        esinE = a_xN * np.sin(Epw) - a_yN * np.cos(Epw)

        e_L = (a_xN**2 + a_yN**2)**(0.5)
        p_L = a * (1 - e_L**2)
        r = a * (1 - ecosE)
        rdot = k_e * np.sqrt(a)/r * esinE
        rfdot = k_e * np.sqrt(p_L) / r
        cosu = a / r * (np.cos(Epw) - a_xN +
                        (a_yN * (esinE) / (1 + np.sqrt(1 - e_L**2))))
        sinu = a / r * (np.sin(Epw) - a_yN +
                        (a_xN * (esinE) / (1 + np.sqrt(1 - e_L**2))))
        u = np.arctan2(sinu, cosu)


        cos2u = np.cos(2*u)
        sin2u = np.sin(2*u)

        Deltar = k_2/(2*p_L) * (1 - theta**2) * cos2u
        Deltau = -k_2/(4*p_L**2) * (7*theta**2 - 1) * sin2u
        DeltaW = 3*k_2 * theta / (2 * p_L**2) * sin2u
        Deltai = 3*k_2 * theta / (2 * p_L**2) * cos2u * np.sin(i_0)
        Deltardot = - k_2 * n / p_L * (1 - theta**2) * sin2u
        Deltarfdot = k_2 * n / p_L * ((1 - theta**2) * cos2u -
                                      3/2.0 * (1 - 3*theta**2))

        # osculating quantities

        r_k = r * (1 - 3/2.0 * k_2 * np.sqrt(1 - e_L**2)/p_L**2 *
                   (3 * theta**2 - 1)) + Deltar
        u_k = u + Deltau
        W_k = W + DeltaW
        i_k = i_0 + Deltai
        rdot_k = rdot + Deltardot
        rfdot_k = rfdot + Deltarfdot

        M_x = -np.sin(W_k) * np.cos(i_k)
        M_y = np.cos(W_k) * np.cos(i_k)
        M_z = np.sin(i_k)

        N_x = np.cos(W_k)
        N_y = np.sin(W_k)
        N_z = 0

        U_x = M_x * np.sin(u_k) + N_x * np.cos(u_k)
        U_y = M_y * np.sin(u_k) + N_y * np.cos(u_k)
        U_z = M_z * np.sin(u_k) + N_z * np.cos(u_k)

        V_x = M_x * np.cos(u_k) - N_x * np.sin(u_k)
        V_y = M_y * np.cos(u_k) - N_y * np.sin(u_k)
        V_z = M_z * np.cos(u_k) - N_z * np.sin(u_k)


        r_x = r_k * U_x
        r_y = r_k * U_y
        r_z = r_k * U_z

        rdot_x = rdot_k * U_x + rfdot_k * V_x
        rdot_y = rdot_k * U_y + rfdot_k * V_y
        rdot_z = rdot_k * U_z + rfdot_k * V_z

        return r_x, r_y, r_z, rdot_x, rdot_y, rdot_z

    def get_latlonalt(self, current_time):
        """Get lon lat and altitude for current time
        """
        pos_x, pos_y, pos_z, vel_x, vel_y, vel_z = \
               self.get_position(current_time)
        del vel_x, vel_y, vel_z
        lon = ((np.arctan2(pos_y * XKMPER, pos_x * XKMPER) - gmst(current_time))
               % (2 * np.pi))

        if lon > np.pi:
            lon -= np.pi * 2
        if lon <= -np.pi:
            lon += np.pi * 2

        r = np.sqrt(pos_x ** 2 + pos_y ** 2)
        lat = np.arctan2(pos_z, r)
        e2 = F * (2 - F)
        while True:
            lat2 = lat
            c = 1/(np.sqrt(1 - e2 * (np.sin(lat2) ** 2)))
            lat = np.arctan2(pos_z + c * e2 *np.sin(lat2), r)
            if abs(lat - lat2) < 1e-10:
                break
        alt = r / np.cos(lat)- c
        alt *= XKMPER
        return lat, lon, alt
# pylint: enable-msg=C0103



def _jdays(current_time):
    """Get the julian day of *current_time*.
    """
    d_t = current_time - datetime.datetime(2000, 1, 1, 12, 0)
    return _days(d_t)

def _days(d_t):
    """Get the days (floating point) from *d_t*.
    """
    return (d_t.days +
            (d_t.seconds +
             d_t.microseconds / (1000000.0)) / (24 * 3600.0))

def gmst(current_time):
    """Greenwich mean sidereal current_time, in radians.
    http://celestrak.com/columns/v02n02/
    """
    now = current_time
    #now = datetime.datetime(1995, 10, 1, 9, 0)
    now0 = datetime.datetime(now.year, now.month, now.day)
    epoch = datetime.datetime(2000, 1, 1, 12, 0)
    du2 = _days(now - epoch)
    d_u = _days(now0 - epoch)

    dus = (du2 - d_u) * 86400
    t_u = d_u / 36525.0
    theta_g_0 = (24110.54841 + t_u * (8640184.812866 +
                                      t_u * (0.093104 - t_u * 6.2 * 10e-6)))
    theta_g = (theta_g_0 + dus * 1.00273790934) % 86400
    return (theta_g / 86400.0) * 2 * np.pi


    
if __name__ == "__main__":
    
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to HRPT level 0 format. Uses AAPP and the aapp1b reader.
"""
import glob
import logging
import os.path
import shutil
import subprocess
import tempfile
from ConfigParser import ConfigParser
import datetime

from mpop.utils import ensure_dir
import mpop.satin.aapp1b
from mpop import CONFIG_PATH

from mpop.satellites import PolarFactory 


WORKING_DIR = "/tmp"

SATPOS_DIR = os.path.sep.join(os.environ["AAPP_PREFIX"].split(os.path.sep)[:-1])
SATPOS_DIR = os.path.join(SATPOS_DIR, "data", "satpos")

LOG = logging.getLogger("hrpt loader")

def get_satpos_file(satpos_time, satname):
    """Return the current satpos file
    """
    satpos_file = os.path.join(SATPOS_DIR,
                               "satpos_"+
                               satname+"_"+
                               satpos_time.strftime("%Y%m%d")+".txt")

    if os.path.exists(satpos_file):
	return satpos_file
    elif satpos_time.hour < 2:
        satpos_time -= datetime.timedelta(days=1)
        satpos_file = os.path.join(SATPOS_DIR,
                                   "satpos_"+
                                   satname+"_"+
                                   satpos_time.strftime("%Y%m%d")+".txt")
        return satpos_file
    else:
        raise IOError("Missing satpos file:" + satpos_file)

def load(satscene):
    """Read data from file and load it into *satscene*.
    """
    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level1",
                                    raw = True):
        options[option] = value
    CASES[satscene.instrument_name](satscene, options)

def load_avhrr(satscene, options):
    """Read avhrr data from file and load it into *satscene*.
    """

    if "filename" not in options:
        raise IOError("No filename given, cannot load.")
    filename = os.path.join(
        options["dir"],
        (satscene.time_slot.strftime(options["filename"])))

    file_list = glob.glob(satscene.time_slot.strftime(filename))

    if len(file_list) > 1:
        raise IOError("More than one hrpt file matching!")
    elif len(file_list) == 0:
        raise IOError("No hrpt file matching!: " +
                      satscene.time_slot.strftime(filename))

    
    filename = file_list[0]

    (handle, tempname) = tempfile.mkstemp(prefix="hrpt_decommuted",
                                          dir=WORKING_DIR)

    os.close(handle)
    del handle
    if(satscene.satname == "metop"):
        satname = "M02"
    else:
        satname = satscene.satname + satscene.number
        
    decommutation(filename, tempname, satscene, options, satname)
    calibration_navigation(tempname, satscene, options)

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    new_dir = conf.get(satscene.instrument_name + "-level2", "dir")
    new_name = conf.get(satscene.instrument_name + "-level2", "filename")
    pathname = os.path.join(new_dir, satscene.time_slot.strftime(new_name))
    LOG.debug("Saving to "+pathname)
    shutil.move(tempname, pathname)
    
    mpop.satin.aapp1b.load(satscene)
    os.remove(pathname)

def convert_to_1b(in_filename, out_filename,
                  time_slot_start, time_slot_end,
                  shortname, orbit):
    """Convert hrpt file to level 1b.
    """
    (handle, tempname) = tempfile.mkstemp(prefix="hrpt_decommuted",
                                          dir=WORKING_DIR)

    os.close(handle)
    del handle
    LOG.debug("Decommuting...")
    decommutation(in_filename, tempname,
                  time_slot_start, time_slot_end,
                  shortname)
    LOG.debug("Calibrating, navigating...")
    calibration_navigation(tempname, time_slot_start, shortname)

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, "regional" + shortname + ".cfg"))
    new_dir = conf.get("avhrr-level2", "dir", raw=True)
    new_name = conf.get("avhrr-level2", "filename", raw=True)
    options = {"satellite": shortname, "orbit": orbit}
    pathname = time_slot_start.strftime(os.path.join(new_dir, new_name))%options
    LOG.debug("Saving to "+pathname)
    ensure_dir(pathname)
    shutil.move(tempname, pathname)

def calibration_navigation(filename, time_slot, shortname):
    """Perform calibration on *filename*
    """
    import pysdh2orbnum
    LOG.info("Calibrating "+filename)
    formated_date = time_slot.strftime("%d/%m/%y %H:%M:%S.000")
    satpos_file = get_satpos_file(time_slot, shortname)
    LOG.debug(formated_date)
    LOG.debug(satpos_file)
    orbit_number = str(pysdh2orbnum.sdh2orbnum(shortname,
                                               formated_date,
                                               satpos_file))
    avhrcl = ("cd /tmp;" +
              "$AAPP_PREFIX/AAPP/bin/avhrcl -c -l -s " +
              shortname + " -d " +
              time_slot.strftime("%Y%m%d") + " -h " +
              time_slot.strftime("%H%M") + " -n " +
              orbit_number + " " +
              filename)

    LOG.debug("Running " + avhrcl)
    proc = subprocess.Popen(avhrcl, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    (out, err) = proc.communicate()
    if out:
        LOG.debug(out)
    if err:
        LOG.error(err)    

    anacl1 = ("cd /tmp;" +
              "$ANA_PATH/bin/ana_lmk_loc -D " + filename)

    LOG.debug("Running " + anacl1)
    proc = subprocess.Popen(anacl1, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    (out, err) = proc.communicate()
    if out:
        LOG.debug(out)
    if err:
        LOG.error(err)    
    
    anacl2 = ("cd /tmp;" +
              "$ANA_PATH/bin/ana_estatt -s " + shortname +
              " -d " + time_slot.strftime("%Y%m%d") +
              " -h " + time_slot.strftime("%H%M") + " -n " +
              orbit_number)

    LOG.debug("Running " + anacl2)
    proc = subprocess.Popen(anacl2, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    (out, err) = proc.communicate()
    if out:
        LOG.debug(out)
    if err:
        LOG.error(err)
        
    avhrcl2 = ("cd /tmp;" +
              "$AAPP_PREFIX/AAPP/bin/avhrcl -l -s " +
              shortname + " -d " +
              time_slot.strftime("%Y%m%d") + " -h " +
              time_slot.strftime("%H%M") + " -n " +
              orbit_number + " " +
              filename)
    LOG.debug("Running " + avhrcl2)
    proc = subprocess.Popen(avhrcl2, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    (out, err) = proc.communicate()
    if out:
        LOG.debug(out)
    if err:
        LOG.error(err)
        
def decommutation(filename_from, filename_to,
                  time_slot_start, time_slot_end,
                  shortname):
    """Perform decommutation on *filename_from* and save the result in
    *filename_to*.
    """
    import pysdh2orbnum
    LOG.info("Decommuting "+filename_from)
    
    

    (handle, tempname) = tempfile.mkstemp(prefix="decommutation",
                                          suffix=".par",
                                          dir=WORKING_DIR)
    os.close(handle)
    handle = open(tempname, "w")
    handle.write("1,0,0,0,0,0,0,0,0,0,0,1\n")
    handle.write("10,11,15,15,16,0,0,13,0,0,0,14\n")
    handle.write(str(time_slot_start.year) + "\n")
    handle.write("0\n")
    
    satpos_file = get_satpos_file(time_slot_start, shortname)
    formated_date = time_slot_start.strftime("%d/%m/%y %H:%M:%S.000")
    orbit_start = str(pysdh2orbnum.sdh2orbnum(shortname,
                                              formated_date,
                                              satpos_file))

    satpos_file = get_satpos_file(time_slot_end, shortname)

    formated_date = time_slot_end.strftime("%d/%m/%y %H:%M:%S.000")
    orbit_end = str(pysdh2orbnum.sdh2orbnum(shortname,
                                            formated_date,
                                            satpos_file))

    handle.write(orbit_start + "," + orbit_end + "\n")
    handle.close()
    
    decom = "$AAPP_PREFIX/AAPP/bin/decommutation"
    cmd = " ".join(["cd " + WORKING_DIR + ";",
                    decom, "ATOVS", tempname, filename_from])
    LOG.debug("Running " + cmd)
    proc = subprocess.Popen(cmd, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    (out, err) = proc.communicate()
    if out:
        LOG.debug(out)
    if err:
        LOG.error(err)

    shutil.move(os.path.join(WORKING_DIR, "hrpt.l1b"), filename_to)
    os.remove(tempname)
    LOG.debug("Decommutation done")

def get_orbit(time_slot, shortname):
    import pysdh2orbnum
    formated_date = time_slot.strftime("%d/%m/%y %H:%M:%S.000")
    
    satpos_file = get_satpos_file(time_slot, shortname)
    
    return str(pysdh2orbnum.sdh2orbnum(shortname,
                                       formated_date,
                                       satpos_file))


def concatenate(granules, channels=None):
    """Concatenate hrpt files.
    """
    filenames = [os.path.join(granule.directory, granule.file_name)
                 for granule in granules]
    
    arg_string = " ".join(filenames)
    
    if filenames[0].endswith(".bz2"):
        cat_cmd = "bzcat"
    else:
        cat_cmd = "cat"


    conffile = os.path.join(CONFIG_PATH, granules[0].fullname + ".cfg")
    conf = ConfigParser()
    conf.read(conffile)
    
    directory = conf.get('avhrr-level1','dir')
    filename = conf.get('avhrr-level1','filename')
    filename = granules[0].time_slot.strftime(filename)
    
    output_name = os.path.join(directory, filename)
    cmd = cat_cmd + " " + arg_string + " > " + output_name
    LOG.debug(cmd)
    proc = subprocess.Popen(cmd, shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    (out, err) = proc.communicate()
    if out:
        LOG.debug(out)
    if err:
        LOG.error(err)

    LOG.debug("Done concatenating level0 files.")

    new_dir = conf.get(granules[0].instrument_name + "-level2", "dir")
    new_name = conf.get(granules[0].instrument_name + "-level2", "filename")
    pathname = os.path.join(new_dir, granules[0].time_slot.strftime(new_name))
    shortname = conf.get('avhrr-level1','shortname')

    convert_to_1b(output_name, pathname,
                  granules[0].time_slot,
                  granules[-1].time_slot + granules[-1].granularity,
                  shortname,
                  get_orbit(granules[0].time_slot,
                            shortname))
    os.remove(output_name)

    scene = PolarFactory.create_scene(granules[0].satname,
                                      granules[0].number,
                                      granules[0].instrument_name,
                                      granules[0].time_slot,
                                      get_orbit(granules[0].time_slot,
                                                shortname),
                                      variant=granules[0].variant)

    scene.load(channels)
    os.remove(pathname)
    return scene

def get_lat_lon(satscene, resolution):
    """Read lat and lon.
    """
    del resolution
    
    return LL_CASES[satscene.instrument_name](satscene, None)

def get_lat_lon_avhrr(satscene, options):
    """Read lat and lon.
    """
    del options
    
    return satscene.lat, satscene.lon

def get_lonlat(satscene, row, col):
    """Read lat and lon.
    """

    return LONLAT_CASES[satscene.instrument_name](satscene, row, col)

def get_lonlat_avhrr(satscene, row, col):
    """Read longitude and latitude for a given pixel.
    """
    # Needs the SATID AAPP env variable to be set to find satid.txt...

    import pyaapp
    import math

    t_start = satscene.time_slot
    epoch = datetime.datetime(1950, 1, 1)
    t50_start = (t_start - epoch)
    jday_start = t50_start.seconds / (3600.0 *24) + t50_start.days
    jday_end = jday_start
    if(satscene.satname == "metop"):
        satname = "M02"
    else:
        satname = satscene.satname + satscene.number

    satpos_file = get_satpos_file(satscene.time_slot, satname)
    
    pyaapp.read_satpos_file(jday_start, jday_end,
                            satscene.satname+" "+str(int(satscene.number)),
                            satpos_file)
    att = pyaapp.prepare_attitude(int(satscene.number), 0, 0, 0)
    lonlat = pyaapp.linepixel2lonlat(int(satscene.number), row, col, att,
                                     jday_start, jday_end)[1:3]
    return (lonlat[0] * 180.0 / math.pi, lonlat[1] * 180.0 / math.pi)

LONLAT_CASES = {
    "avhrr": get_lonlat_avhrr
    }


LL_CASES = {
    "avhrr": get_lat_lon_avhrr
    }

CASES = {
    "avhrr": load_avhrr
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2015, 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>
#   Katerina.Melnik <kmelnik@scanex.ru>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A VIRR reader for FY3-B and maybe A....
"""
import numpy as np
import os
import logging
from datetime import datetime
from ConfigParser import ConfigParser
from mpop import CONFIG_PATH
import h5py
from pyspectral.blackbody import blackbody_wn_rad2temp as rad2temp

LOGGER = logging.getLogger('virr')


def load(satscene, *args, **kwargs):
    """Read data from file and load it into *satscene*.
    A possible *calibrate* keyword argument is passed to the AAPP reader.
    Should be 0 for off (counts), 1 for default (brightness temperatures and
    reflectances), and 2 for radiances only.

    If *use_extern_calib* keyword argument is set True, use external
    calibration data.

    """
    del args

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level2",
                                    raw=True):
        options[option] = value

    if kwargs.get("filename") is not None:
        options["full_filename"] = kwargs["filename"]
    if kwargs.get("calibrate") is not None:
        options["calibrate"] = kwargs["calibrate"]
    else:
        options["calibrate"] = True

    LOGGER.debug("Calibrate = " + str(options["calibrate"]))
    LOGGER.info("Loading instrument '%s'", satscene.instrument_name)

    try:
        CASES[satscene.instrument_name](satscene, options)
    except KeyError:
        raise KeyError("Unknown instrument '%s'" % satscene.instrument_name)


def load_virr(satscene, options):
    """Read the VIRR hdf5 file"""

    if "filename" not in options:
        raise IOError("No 1km virr filename given, cannot load")

    values = {"orbit": satscene.orbit,
              "satname": satscene.satname,
              "instrument": satscene.instrument_name,
              "satellite": satscene.fullname
              }

    filename = \
        os.path.join(satscene.time_slot.strftime(options["dir"]) % values,
                     satscene.time_slot.strftime(
                         options["filename"])
                     % values)

    LOGGER.debug("Filename= %s", filename)

    datasets = ['EV_Emissive',
                'EV_RefSB']

    calibrate = options['calibrate']
    LOGGER.debug("Calibrate = " + str(calibrate))

    h5f = h5py.File(filename, 'r')

    # Get geolocation information
    lons = h5f['Longitude'][:]
    lats = h5f['Latitude'][:]
    # Mask out unrealistic values:
    mask = np.logical_or(lats > 90., lons > 90.)
    lons = np.ma.masked_array(lons, mask=mask)
    lats = np.ma.masked_array(lats, mask=mask)
    sunz = h5f['SolarZenith'][:]
    slope = h5f['SolarZenith'].attrs['Slope'][0]
    intercept = h5f['SolarZenith'].attrs['Intercept'][0]
    sunz = sunz * slope + intercept
    sunz = np.where(np.greater(sunz, 85.0), 85.0, sunz)

    # Get the calibration information
    # Emissive radiance coefficients:
    emis_offs = h5f['Emissive_Radiance_Offsets'][:]
    emis_scales = h5f['Emissive_Radiance_Scales'][:]

    # Central wave number (unit =  cm-1) for the three IR bands
    # It is ordered according to decreasing wave number (increasing wavelength):
    # 3.7 micron, 10.8 micron, 12 micron
    emiss_centroid_wn = h5f.attrs['Emmisive_Centroid_Wave_Number']

    # VIS/NIR calibration stuff:
    refsb_cal_coeff = h5f.attrs['RefSB_Cal_Coefficients']
    visnir_scales = refsb_cal_coeff[0::2]
    visnir_offs = refsb_cal_coeff[1::2]

    refsb_effective_wl = h5f.attrs['RefSB_Effective_Wavelength']

    # Read the band data:
    for dset in datasets:
        band_data = h5f[dset]
        valid_range = band_data.attrs['valid_range']
        LOGGER.debug("valid-range = " + str(valid_range))
        fillvalue = band_data.attrs['_FillValue']
        band_names = band_data.attrs['band_name'].split(',')
        slope = band_data.attrs['Slope']
        intercept = band_data.attrs['Intercept']
        units = band_data.attrs['units']
        long_name = band_data.attrs['long_name']

        LOGGER.debug('band names = ' + str(band_names))

        for (i, band) in enumerate(band_names):
            if band not in satscene.channels_to_load:
                continue

            LOGGER.debug("Reading channel %s, i=%d", band, i)
            data = band_data[i]

            bandmask = np.logical_or(np.less(data, valid_range[0]),
                                     np.greater(data, valid_range[1]))

            if calibrate:
                if dset in ['EV_Emissive']:
                    data = (np.array([emis_offs[:, i]]).transpose() +
                            data * np.array([emis_scales[:, i]]).transpose())
                    # Radiance to Tb conversion.
                    # Pyspectral wants SI units,
                    # but radiance data are in mW/m^2/str/cm^-1 and wavenumbers are in cm^-1
                    # Therefore multply wavenumber by 100 and radiances by
                    # 10^-5
                    data = rad2temp(emiss_centroid_wn[i] * 100., data * 1e-5)
                    LOGGER.debug("IR data calibrated")

                if dset in ['EV_RefSB']:
                    data = (visnir_offs[i] +
                            data * visnir_scales[i]) / np.cos(np.deg2rad(sunz))

            satscene[band] = np.ma.masked_array(data,
                                                mask=bandmask,
                                                copy=False)

    from pyresample import geometry
    satscene.area = geometry.SwathDefinition(lons=lons, lats=lats)

    h5f.close()


CASES = {
    "virr": load_virr,
}
